from datetime import datetime
import random
import string

from restic_replica.snapshots import Snapshot


def new_snapshot(timestamp: datetime):
    id = "".join(
        random.choice(string.ascii_lowercase + string.digits) for _ in range(64)
    )
    short_id = id[:8]
    return Snapshot(
        timestamp,
        "a9c65ce7565f9e7456606dd0119ab186ba5aefc6fb883f433e7a6b406c0f6771",
        ["/etc/hosts"],
        "server.local",
        "user",
        1000,
        1000,
        "restic 0.16.4",
        id,
        short_id,
    )



pytest_plugins = [
    "tests.fixtures.logger_fixture",
    "tests.fixtures.repository_fixture",
    "tests.fixtures.restic_cli_fixture",
    "tests.fixtures.snapshot_fixture",
    "tests.fixtures.snapshot_list_fixture",
]



from datetime import datetime
import logging
import pytest
from unittest import mock
from pathlib import Path

from restic_replica import __version__, console


class TestInfoOnly:
    """Tests for the class console.InfoOnly"""

    @pytest.mark.parametrize(
        "log_level, expectation",
        [
            (logging.DEBUG, False),
            (logging.INFO, True),
            (logging.WARNING, False),
            (logging.ERROR, False),
            (logging.CRITICAL, False),
        ],
    )
    def test_filter(self, log_level, expectation):
        """Function filter should return only messages of level logging.INFO"""

        log_record = logging.LogRecord(None, log_level, None, None, None, None, None)
        assert console.InfoOnly.filter(None, log_record) is expectation


class TestNoInfo:
    """Tests for the class console.NoInfo"""

    @pytest.mark.parametrize(
        "log_level, expectation",
        [
            (logging.DEBUG, True),
            (logging.INFO, False),
            (logging.WARNING, True),
            (logging.ERROR, True),
            (logging.CRITICAL, True),
        ],
    )
    def test_filter(self, log_level, expectation):
        """Function filter should return all messages except those with level logging.INFO"""

        log_record = logging.LogRecord(None, log_level, None, None, None, None, None)
        assert console.NoInfo.filter(None, log_record) is expectation


class TestParseCliArgs:
    """Tests for the function console.parse_cli_args"""

    def test_version(self, capsys):
        """Should print the name and version of the package, then exit"""
        with pytest.raises(SystemExit):
            console.parse_cli_args(["--version"])
        captured = capsys.readouterr()
        assert captured.out.strip() == f"restic-replica {__version__}"

    def test_verbose(self):
        """should set verbosity level"""
        assert console.parse_cli_args([]).verbose == 0
        assert console.parse_cli_args(["-v"]).verbose == 1
        assert console.parse_cli_args(["--verbose"]).verbose == 1

    def test_dry_run(self):
        """should set dry_run boolean"""
        assert console.parse_cli_args([]).dry_run is False
        assert console.parse_cli_args(["--dry-run"]).dry_run is True


class TestSetupLogging:
    """Tests for the function console.setup_logging"""

    @pytest.mark.usefixtures("logger_fixture")
    def test_handlers(self, logger_fixture, tmp_path):
        """Two StreamHandlers and one FileHandler should be configured"""
        logger = console.setup_logging(logger_fixture, logdir=tmp_path)
        assert len(logger.handlers) == 3
        assert isinstance(logger.handlers[0], logging.StreamHandler)
        assert isinstance(logger.handlers[1], logging.StreamHandler)
        assert isinstance(logger.handlers[2], logging.FileHandler)

    @pytest.mark.usefixtures("logger_fixture")
    def test_custom_logdir(self, logger_fixture, tmp_path):
        """Custom logdir should be set correctly"""
        logger = console.setup_logging(logger_fixture, logdir=tmp_path)
        assert Path(logger.handlers[2].baseFilename).parent == tmp_path

    @pytest.mark.usefixtures("logger_fixture")
    def test_no_logdir(self, logger_fixture):
        """Logdir set to None should result in no logfile handler"""
        logger = console.setup_logging(logger_fixture, logdir=None)
        assert len(logger.handlers) == 2
        assert isinstance(logger.handlers[0], logging.StreamHandler)
        assert isinstance(logger.handlers[1], logging.StreamHandler)

    @pytest.mark.usefixtures("logger_fixture")
    def test_logname(self, logger_fixture, tmp_path):
        """Log name should be made up of program name and iso timestamp"""
        logger = console.setup_logging(logger_fixture, logdir=tmp_path)
        logname = Path(logger.handlers[2].baseFilename).stem
        logname_parts = logname.split("_")
        assert logname_parts[0] == "restic-replica"
        assert datetime.fromisoformat(logname_parts[1]) is not None

    @pytest.mark.usefixtures("logger_fixture")
    @pytest.mark.parametrize("debug", [True, False])
    def test_debug(self, logger_fixture, tmp_path, debug):
        """Setting debug boolen should change the logger level"""
        logger = console.setup_logging(logger_fixture, logdir=tmp_path, debug=debug)
        assert logger.getEffectiveLevel() == logging.DEBUG if debug else logging.INFO

    @pytest.mark.usefixtures("logger_fixture")
    def test_logdir_error(self, logger_fixture, tmp_path):
        """Inaccessible log directory should raise an OSError"""
        with mock.patch("pathlib.Path.mkdir", side_effect=PermissionError):
            with pytest.raises(OSError):
                console.setup_logging(logger_fixture, logdir=tmp_path)


class TestLoggingHeaders:
    """Tests for the function console.logging_headers"""

    def test_logging_headers(self, caplog):
        """Logging headers should be emitted"""
        caplog.set_level(logging.INFO)
        console.logging_headers("9.9.9")
        ts = (
            caplog.messages[3].split("@")[1].strip()
        )  # extract timestamp from the 4th line
        assert caplog.messages == [
            "==============================",
            "  restic-replica 9.9.9",
            "==============================",
            f"Program start @ {ts}",
        ]



import sys

from restic_replica import __version__, app, console


def main(argv=sys.argv[1:]):
    args = console.parse_cli_args(argv)
    config_file = app.ensure_config_file()
    config = app.read_config_file(config_file)
    logger = console.setup_logging(logdir=app.get_logdir(config))
    console.logging_headers(__version__)
    try:
        # get restic cli
        restic_cli = app.get_restic(config["restic"], verbose=args.verbose)
        # read policy info from file
        policy = app.get_policy(config["policy"])
        # instance source and target repositories
        source = app.get_repository("source", config["source"], restic_cli)
        destination = app.get_repository("target", config["destination"], restic_cli)
        # check access, then trigger copy
        logger.info(f"Checking access to source repository: {source.uri}")
        app.check_repository_access(source)
        logger.info(f"Checking access to destination repository: {destination.uri}")
        app.check_repository_access(destination)
        # perform snapshot copy operation
        result = app.copy_snapshots(source, destination, policy, dry_run=args.dry_run)
        if not result.stdout:
            logger.info(
                "All snapshots from the source are already present in the destination repository"
            )
        else:
            logger.info("Finished copying snapshots")
    except RuntimeError as err:
        logger.error(err)
        raise SystemExit(1)


if __name__ == "__main__":  # pragma: no cover
    main()



import argparse
from datetime import datetime
import logging
from pathlib import Path
from typing import Optional

from restic_replica import __version__

logger = logging.getLogger(__name__)


class InfoOnly(logging.Filter):
    """Return only messages of level logging.INFO"""

    def filter(self, record):
        return record.levelno == logging.INFO


class NoInfo(logging.Filter):
    """Return all messages except those with level logging.INFO"""

    def filter(self, record):
        return record.levelno != logging.INFO


def parse_cli_args(argv: Optional[list[str]] = None) -> argparse.Namespace:
    """
    Process arguments provided via cli

    Returns:
        args: Simple object used for storing key/value attribute pairs
    """
    about = "Copy snapshots from one restic repository to another"
    parser = argparse.ArgumentParser(prog="restic-replica", description=about)
    parser.add_argument(
        "--version", action="version", version=f"%(prog)s {__version__}"
    )
    parser.add_argument(
        "-v",
        "--verbose",
        help="make restic output more verbose (specify multiple times, max level/times is 2)",
        action="count",
        default=0,
    )
    parser.add_argument(
        "--dry-run",
        help="do not actually copy any snapshots, print what would be done",
        action="store_true",
    )
    return parser.parse_args(argv)


def setup_logging(
    logger: logging.Logger = logging.getLogger("restic_replica"),
    logdir: Optional[Path] = None,
    debug: bool = False,
) -> logging.Logger:
    """
    Setup the application logger, a logging.Logger instance. Configures logging to
    stdout, and optionally to file.

    Args:
        logger: the logging.Logger instance to configure.
        logdir: an optional path to a directory where a log file will be written.
        debug: boolean that sets the logging level to debug when True

    Returns:
        the configured Logger instance.
    """

    # set debugging log level if specified
    if debug:
        logger.setLevel(logging.DEBUG)
    else:
        logger.setLevel(logging.INFO)

    # console handling is split into two, so the levelname is not prepended to info messages, for aesthetic reasons.
    # setup logging to console for all messages _except_ info (i.e. debug, warn, error, fatal)
    ch = logging.StreamHandler()
    ch.addFilter(NoInfo())
    ch.setFormatter(logging.Formatter(fmt="%(levelname)s: %(message)s"))
    logger.addHandler(ch)
    # setup logging to console for info messages
    ch_info = logging.StreamHandler()
    ch_info.addFilter(InfoOnly())
    logger.addHandler(ch_info)

    # setup logging to file
    if logdir:
        timestamp = datetime.now().strftime("%Y%m%dT%H%M%S")
        logname = f"restic-replica_{timestamp}.log"
        # create logging dir
        logdir.mkdir(parents=True, exist_ok=True)
        fh = logging.FileHandler(logdir / f"{logname}")
        fh_formatter = logging.Formatter(
            fmt="%(asctime)s %(levelname)s: %(message)s",
            datefmt="%Y-%m-%d %H:%M:%S%z",
        )
        fh.setFormatter(fh_formatter)
        logger.addHandler(fh)

    # return the logger object
    return logger


def logging_headers(version: str) -> None:
    """
    Write the application headers to the application logger.

    Args:
        version: The application version.
    """
    logger.info("==============================")
    logger.info(f"  restic-replica {version}")
    logger.info("==============================")
    logger.info(f"Program start @ {datetime.now().strftime("%Y/%m/%d %H:%M:%S%z")}")



import pytest
import random
import string

import logging


@pytest.fixture
def logger_fixture():
    """return a (class:`logging.logger`) instance"""
    name = "".join(random.choices(string.ascii_letters + string.digits, k=10))
    return logging.getLogger(name)



import datetime
import pytest

from restic_replica.snapshots import Snapshot


@pytest.fixture
def snapshot_fixture():
    """return a (class:`snapshots.Snapshot`) instance"""
    return Snapshot(
        datetime.datetime(
            2025,
            9,
            22,
            15,
            19,
            14,
            968650,
            tzinfo=datetime.timezone(datetime.timedelta(seconds=3600)),
        ),
        "a9c65ce7565f9e7456606dd0119ab186ba5aefc6fb883f433e7a6b406c0f6771",
        ["/etc/hosts"],
        "server.local",
        "user",
        1000,
        1000,
        "restic 0.18.0",
        "13fc6fb1a3ce4ba6a693bc7e0f6f651394e0699db4c38080c2f7c1fabe5210b2",
        "13fc6fb1",
        parent="ef699e0b81670666e639c0271b09edc6b4e3158277e3dd1c0d72809b44c468f1",
        original="e2adfd3564420f9447d42337356100a168dbf9c1de25b3086fbdc9c4a18ba4a1",
        excludes=["/etc/nothosts"],
        tags=["rewrite"],
        summary={
            "backup_start": "2025-09-22T15:19:14.968650111+01:00",
            "backup_end": "2025-09-22T15:19:15.693199959+01:00",
            "files_new": 1,
            "files_changed": 0,
            "files_unmodified": 0,
            "dirs_new": 1,
            "dirs_changed": 0,
            "dirs_unmodified": 0,
            "data_blobs": 1,
            "tree_blobs": 2,
            "data_added": 1288,
            "data_added_packed": 1028,
            "total_files_processed": 1,
            "total_bytes_processed": 384,
        },
    )



from pathlib import Path
import pytest

from restic_replica.repository import ResticCli


@pytest.fixture
def restic_cli_fixture():
    """Return a (:class:`repository.ResticCli`) instance"""
    return ResticCli(
        Path("restic"), environment_vars={"RESTIC_PROGRESS_FPS": "0.016667"}
    )



from contextlib import nullcontext as does_not_raise
import datetime
import pytest
import random
import textwrap
from unittest import mock

from restic_replica import snapshots
from tests.utils import new_snapshot


class TestPolicy:
    """Tests for the class repository.Policy"""

    class TestPostInit:
        """Tests for the __post_init__ method"""

        @pytest.mark.parametrize(
            "arg, expectation",
            [
                (1, does_not_raise()),
                (0, pytest.raises(ValueError)),
                (-1, pytest.raises(ValueError)),
                ("foo", pytest.raises(TypeError)),
            ],
        )
        def test_positive_int_only(self, arg, expectation):
            with expectation:
                snapshots.Policy(arg, 0, 0, 0, 0)
                snapshots.Policy(0, arg, 0, 0, 0)
                snapshots.Policy(0, 0, arg, 0, 0)
                snapshots.Policy(0, 0, 0, arg, 0)
                snapshots.Policy(0, 0, 0, 0, arg)

        @pytest.mark.parametrize(
            "arg, expectation",
            [
                (True, does_not_raise()),
                (False, does_not_raise()),
                (0, does_not_raise()),  # int 0 maps to False
                (1, does_not_raise()),  # int 1 maps to True
                (2, pytest.raises(TypeError)),  # int gt 1 does not resolve to bool
                (-1, pytest.raises(TypeError)),  # int lt 0 does not resolve to bool
                ("foo", pytest.raises(TypeError)),
            ],
        )
        def test_no_current(self, arg, expectation):
            """exclude-current-period should only accept a boolean (or a 0, or a 1, as those resolve to true/false)"""
            with expectation:
                snapshots.Policy(1, no_current=arg)

    class TestStr:
        """Tests for the __str__ method"""

        @pytest.mark.parametrize(
            "policy, expectation",
            [
                (snapshots.Policy(5, 0, 0, 0, 0), "keep-last=5"),
                (snapshots.Policy(0, 4, 0, 0, 0), "keep-daily=4"),
                (snapshots.Policy(0, 0, 3, 0, 0), "keep-weekly=3"),
                (snapshots.Policy(0, 0, 0, 2, 0), "keep-monthly=2"),
                (snapshots.Policy(0, 0, 0, 0, 1), "keep-yearly=1"),
                (
                    snapshots.Policy(5, 4, 3, 2, 1),
                    "keep-last=5, keep-daily=4, keep-weekly=3, keep-monthly=2, keep-yearly=1",
                ),
                (
                    snapshots.Policy(5, 0, 0, 0, 0, True),
                    "keep-last=5, exclude-current-period=True",
                ),
            ],
        )
        def test_output(self, policy, expectation):
            assert str(policy) == expectation


class TestSnapshot:
    """Tests for the class snapshots.Snapshot"""

    class TestFromDict:
        """Tests for the from_dict method"""

        def test_no_optionals(self):
            """valid snapshot should be returned if no optional parameters are specified"""
            data = {
                "time": "2025-09-22T15:19:14.968650111+01:00",
                "tree": "a9c65ce7565f9e7456606dd0119ab186ba5aefc6fb883f433e7a6b406c0f6771",
                "paths": ["/etc/hosts"],
                "hostname": "server.local",
                "username": "user",
                "uid": 1000,
                "gid": 1000,
                "program_version": "restic 0.18.0",
                "id": "13fc6fb1a3ce4ba6a693bc7e0f6f651394e0699db4c38080c2f7c1fabe5210b2",
                "short_id": "13fc6fb1",
            }

            assert snapshots.Snapshot.from_dict(data) == snapshots.Snapshot(
                datetime.datetime(
                    2025,
                    9,
                    22,
                    15,
                    19,
                    14,
                    968650,
                    tzinfo=datetime.timezone(datetime.timedelta(seconds=3600)),
                ),
                "a9c65ce7565f9e7456606dd0119ab186ba5aefc6fb883f433e7a6b406c0f6771",
                ["/etc/hosts"],
                "server.local",
                "user",
                1000,
                1000,
                "restic 0.18.0",
                "13fc6fb1a3ce4ba6a693bc7e0f6f651394e0699db4c38080c2f7c1fabe5210b2",
                "13fc6fb1",
            )

        @pytest.mark.usefixtures("snapshot_fixture")
        def test_with_optionals(self, snapshot_fixture):
            """valid snapshot should be returned if any optional parameters are specified"""
            data = {
                "time": "2025-09-22T15:19:14.968650111+01:00",
                "parent": "ef699e0b81670666e639c0271b09edc6b4e3158277e3dd1c0d72809b44c468f1",
                "tree": "a9c65ce7565f9e7456606dd0119ab186ba5aefc6fb883f433e7a6b406c0f6771",
                "paths": ["/etc/hosts"],
                "hostname": "server.local",
                "username": "user",
                "uid": 1000,
                "gid": 1000,
                "excludes": ["/etc/nothosts"],
                "tags": ["rewrite"],
                "original": "e2adfd3564420f9447d42337356100a168dbf9c1de25b3086fbdc9c4a18ba4a1",
                "program_version": "restic 0.18.0",
                "summary": {
                    "backup_start": "2025-09-22T15:19:14.968650111+01:00",
                    "backup_end": "2025-09-22T15:19:15.693199959+01:00",
                    "files_new": 1,
                    "files_changed": 0,
                    "files_unmodified": 0,
                    "dirs_new": 1,
                    "dirs_changed": 0,
                    "dirs_unmodified": 0,
                    "data_blobs": 1,
                    "tree_blobs": 2,
                    "data_added": 1288,
                    "data_added_packed": 1028,
                    "total_files_processed": 1,
                    "total_bytes_processed": 384,
                },
                "id": "13fc6fb1a3ce4ba6a693bc7e0f6f651394e0699db4c38080c2f7c1fabe5210b2",
                "short_id": "13fc6fb1",
            }
            assert snapshots.Snapshot.from_dict(data) == snapshot_fixture


class TestSnapshotList:
    """Tests for the class snapshots.SnapshotList"""

    class TestStr:
        """Tests for the __str__ method"""

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_output(self, snapshot_list_fixture):
            """output should be a list of snapshot short_ids"""
            assert [x.strip() for x in str(snapshot_list_fixture).split(",")] == [
                s.short_id for s in snapshot_list_fixture.snapshots
            ]

    class TestFromJson:
        """Tests for the from_json method"""

        @pytest.mark.usefixtures("snapshot_fixture")
        def test_ingest(self, snapshot_fixture):
            """json should be parsed and return valid snapshotlist and snapshot instances"""
            data = textwrap.dedent(
                # fmt: off
                "["
                    "{"
                        '"time":"2025-09-22T15:19:14.968650111+01:00",'
                        '"parent": "ef699e0b81670666e639c0271b09edc6b4e3158277e3dd1c0d72809b44c468f1",'
                        '"tree":"a9c65ce7565f9e7456606dd0119ab186ba5aefc6fb883f433e7a6b406c0f6771",'
                        '"paths":["/etc/hosts"],'
                        '"hostname":"server.local",'
                        '"username":"user",'
                        '"uid":1000,'
                        '"gid":1000,'
                        '"excludes": ["/etc/nothosts"],'
                        '"tags": ["rewrite"],'
                        '"original": "e2adfd3564420f9447d42337356100a168dbf9c1de25b3086fbdc9c4a18ba4a1",'
                        '"program_version":"restic 0.18.0",'
                        '"summary": {'
                            '"backup_start": "2025-09-22T15:19:14.968650111+01:00",'
                            '"backup_end": "2025-09-22T15:19:15.693199959+01:00",'
                            '"files_new": 1,'
                            '"files_changed": 0,'
                            '"files_unmodified": 0,'
                            '"dirs_new": 1,'
                            '"dirs_changed": 0,'
                            '"dirs_unmodified": 0,'
                            '"data_blobs": 1,'
                            '"tree_blobs": 2,'
                            '"data_added": 1288,'
                            '"data_added_packed": 1028,'
                            '"total_files_processed": 1,'
                            '"total_bytes_processed": 384'
                        "},"
                        '"id":"13fc6fb1a3ce4ba6a693bc7e0f6f651394e0699db4c38080c2f7c1fabe5210b2",'
                        '"short_id":"13fc6fb1"'
                    "}"
                "]"
                # fmt: on
            )
            assert snapshots.SnapshotList.from_json(data).snapshots == [
                snapshot_fixture
            ]

        def test_null_ingest(self):
            """empty json data should return an empty snapshotlist instance"""
            data = "[]"
            assert snapshots.SnapshotList.from_json(data).snapshots == []

    class TestTimeSorted:
        """Tests for the time_sorted method"""

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_sort_ascending(self, snapshot_list_fixture):
            """snapshots should be sorted by date ascending"""
            random.shuffle(snapshot_list_fixture.snapshots)
            sorted = snapshot_list_fixture.time_sorted(descending=False)
            for i, snap in enumerate(sorted):
                if i != 0:
                    assert snap.time > sorted[i - 1].time

        def test_sort_descending(self, snapshot_list_fixture):
            """snapshots should be sorted by date descending"""
            random.shuffle(snapshot_list_fixture.snapshots)
            sorted = snapshot_list_fixture.time_sorted(descending=True)
            for i, snap in enumerate(sorted):
                if i != 0:
                    assert snap.time < sorted[i - 1].time

    class TestFilter:
        """Tests for the filter method"""

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_no_snapshots(self, snapshot_list_fixture):
            """if no filters return any snapshots, then no snapshots should be returned"""
            # snapshot_list_fixture has 10 snapshots
            # set a policy that will return nothing
            policy = snapshots.Policy(1, 0, 0, 0, 0)
            # set "last" to zero after init, to avoid triggering "all zeroes not allowed" check
            policy.last = 0
            assert snapshot_list_fixture.filter(policy) == []

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_first_sort(self, snapshot_list_fixture):
            """snapshots should be sorted into descending order of time before being passed to _filter* functions"""
            # policy.last set to non-zero value, so _filter_last will trigger
            policy = snapshots.Policy(99999, 0, 0, 0, 0)
            with mock.patch.object(
                snapshot_list_fixture, "_filter_last", return_value=[]
            ) as target:
                snapshot_list_fixture.filter(policy)
            # assert that filter_last was called with the snapshots list sorted in reverse
            target.assert_called_with(
                policy.last, list(reversed(snapshot_list_fixture.snapshots))
            )

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_return_sort(self, snapshot_list_fixture):
            """snapshots should be sorted into ascending order when returned"""
            # invert the snapshot list, so it is sorted in reverse
            snap_list = list(reversed(snapshot_list_fixture.snapshots))
            # policy.last set to non-zero value, so _filter_last will trigger
            policy = snapshots.Policy(1, 0, 0, 0, 0)
            # filter_last mocked to return reversed snapshot list
            with mock.patch.object(
                snapshot_list_fixture, "_filter_last", return_value=snap_list
            ):
                # output should be sorted into ascending order again
                assert (
                    snapshot_list_fixture.filter(policy)
                    == snapshot_list_fixture.snapshots
                )

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_filters_merge(self, snapshot_list_fixture, monkeypatch):
            """snapshots selected by each filter should all be combined into the output"""

            def new_snaps(*args, **kwargs):
                return [new_snapshot(datetime.date(2025, 9, 29))]

            # enable all filters, by setting to non-zero value
            policy = snapshots.Policy(1, 1, 1, 1, 1)
            # mock all filters to return unqiue snapshots
            monkeypatch.setattr(snapshot_list_fixture, "_filter_last", new_snaps)
            monkeypatch.setattr(snapshot_list_fixture, "_filter_daily", new_snaps)
            monkeypatch.setattr(snapshot_list_fixture, "_filter_weekly", new_snaps)
            monkeypatch.setattr(snapshot_list_fixture, "_filter_monthly", new_snaps)
            monkeypatch.setattr(snapshot_list_fixture, "_filter_yearly", new_snaps)
            # verify 5 snapshots are in the output
            assert len(snapshot_list_fixture.filter(policy)) == 5

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_no_duplicates(self, snapshot_list_fixture):
            """snapshots which are selected by multiple filters should only appear once in the output"""
            # each filter will select one snapshot (the most recent)
            policy = snapshots.Policy(1, 1, 1, 1, 1)
            expected = snapshot_list_fixture.snapshots[-1]
            assert snapshot_list_fixture.filter(policy) == [expected]

    class TestFilterLast:
        """Tests for the _filter_last method"""

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_no_snapshots(self, snapshot_list_fixture):
            """an empty list of snapshots should return an empty list"""
            assert snapshot_list_fixture._filter_last(5, []) == []

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_some_snapshots(self, snapshot_list_fixture):
            """the requested number of snapshots should be returned, from the top of the list"""
            snap_list = snapshot_list_fixture.snapshots
            assert (
                snapshot_list_fixture._filter_last(3, snap_list)
                == snapshot_list_fixture.snapshots[:3]
            )

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_not_enough_snapshots(self, snapshot_list_fixture):
            """if more than the total number of snapshots are requested, the entire list should be returned"""
            # snapshot_list_fixture has 10 total snapshots
            snap_list = snapshot_list_fixture.snapshots
            assert len(snapshot_list_fixture._filter_last(20, snap_list)) == 10

    class TestFilterDaily:
        """Tests for the _filter_daily method"""

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_zero(self, snapshot_list_fixture):
            """requesting zero snapshots should return an empty list"""
            snap_list = snapshot_list_fixture.snapshots
            assert snapshot_list_fixture._filter_daily(0, snap_list) == []

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_sequential_days(self, snapshot_list_fixture):
            """snapshots on sequential days should be returned up to the requested number"""
            snap_list = [
                new_snapshot(datetime.date(2025, 9, 21)),
                new_snapshot(datetime.date(2025, 9, 20)),
                new_snapshot(datetime.date(2025, 9, 19)),
                new_snapshot(datetime.date(2025, 9, 18)),
            ]
            assert snapshot_list_fixture._filter_daily(3, snap_list) == snap_list[:3]

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_non_sequential_days(self, snapshot_list_fixture):
            """snapshots on non-sequential days should be returned up to the requested number"""
            snap_list = [
                new_snapshot(datetime.date(2025, 9, 21)),
                new_snapshot(datetime.date(2025, 9, 19)),
                new_snapshot(datetime.date(2025, 9, 18)),
                new_snapshot(datetime.date(2025, 9, 17)),
            ]
            assert snapshot_list_fixture._filter_daily(3, snap_list) == snap_list[:3]

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_max_one_per_day(self, snapshot_list_fixture):
            """only the most recent snapshot on any given calendar day should be considered"""
            snap_list = [
                new_snapshot(datetime.datetime(2025, 9, 21, 15, 19, 14)),
                new_snapshot(datetime.datetime(2025, 9, 21, 10, 34, 9)),
                new_snapshot(datetime.datetime(2025, 9, 21, 7, 1, 42)),
                new_snapshot(datetime.date(2025, 9, 20)),
                new_snapshot(datetime.date(2025, 9, 19)),
            ]
            assert snapshot_list_fixture._filter_daily(3, snap_list) == [
                snap_list[0],
                snap_list[3],
                snap_list[4],
            ]

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_week_boundary(self, snapshot_list_fixture):
            """snapshots should be able to span the week boundary without issue"""
            snap_list = [
                new_snapshot(datetime.date(2025, 9, 23)),
                new_snapshot(datetime.date(2025, 9, 22)),  # monday
                new_snapshot(datetime.date(2025, 9, 21)),  # sunday
                new_snapshot(datetime.date(2025, 9, 20)),
            ]
            assert snapshot_list_fixture._filter_daily(3, snap_list) == snap_list[:3]

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_month_boundary(self, snapshot_list_fixture):
            """snapshots should be able to span the month boundary without issue"""
            snap_list = [
                new_snapshot(datetime.date(2025, 9, 23)),
                new_snapshot(datetime.date(2025, 9, 22)),
                new_snapshot(datetime.date(2025, 8, 23)),
                new_snapshot(datetime.date(2025, 8, 22)),
            ]
            assert snapshot_list_fixture._filter_daily(3, snap_list) == snap_list[:3]

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_year_boundary(self, snapshot_list_fixture):
            """snapshots should be able to span the year boundary without issue"""
            snap_list = [
                new_snapshot(datetime.date(2025, 9, 23)),
                new_snapshot(datetime.date(2025, 9, 22)),
                new_snapshot(datetime.date(2024, 9, 23)),
                new_snapshot(datetime.date(2024, 9, 22)),
            ]
            assert snapshot_list_fixture._filter_daily(3, snap_list) == snap_list[:3]

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_not_enough_snapshots(self, snapshot_list_fixture):
            """if more than the total number of daily snapshots are requested, all daily snapshots should be returned"""
            # snapshot_list_fixture has three snapshots from a single day, out of 10 total
            snap_list = snapshot_list_fixture.snapshots
            assert len(snapshot_list_fixture._filter_daily(10, snap_list)) == 8

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_no_current(self, snapshot_list_fixture):
            """if no_current is set and a snapshot lies within the current time period, it should be excluded"""
            snap_list = [
                new_snapshot(datetime.datetime(2025, 9, 21)),
                new_snapshot(datetime.datetime(2025, 9, 20)),
                new_snapshot(datetime.datetime(2025, 9, 19)),
                new_snapshot(datetime.datetime(2025, 9, 18)),
            ]
            assert (
                snapshot_list_fixture._filter_daily(
                    3,
                    snap_list,
                    no_current=True,
                    current_time=datetime.datetime(2025, 9, 21),
                )
                == snap_list[1:]
            )

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_snapshot_timezone_primacy(self, snapshot_list_fixture):
            """
            the timezone recorded in the snapshot should have primacy when determining in which period the snapshot falls
            e.g. 2018-11-03T01:38:41+01:00, and 2018-11-04T01:38:41+06:00, should be considered as having been taken on different days.
            this matches restic behaviour: https://forum.restic.net/t/which-timezone-does-forget-use/1166
            """
            snap_list = [
                new_snapshot(
                    datetime.datetime(
                        2018,
                        11,
                        4,
                        1,
                        38,
                        41,
                        tzinfo=datetime.timezone(datetime.timedelta(seconds=21600)),
                    )
                ),
                new_snapshot(
                    datetime.datetime(
                        2018,
                        11,
                        3,
                        1,
                        38,
                        41,
                        tzinfo=datetime.timezone(datetime.timedelta(seconds=3600)),
                    )
                ),
            ]
            assert (
                len(snapshot_list_fixture._filter_daily(2, snap_list, no_current=True))
                == 2
            )

    class TestFilterWeekly:
        """Tests for the _filter_weekly method"""

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_zero(self, snapshot_list_fixture):
            """requesting zero snapshots should return an empty list"""
            snap_list = snapshot_list_fixture.snapshots
            assert snapshot_list_fixture._filter_weekly(0, snap_list) == []

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_sequential_weeks(self, snapshot_list_fixture):
            """snapshots on sequential weeks should be returned up to the requested number"""
            snap_list = [
                new_snapshot(datetime.date(2025, 9, 29)),
                new_snapshot(datetime.date(2025, 9, 22)),
                new_snapshot(datetime.date(2025, 9, 15)),
                new_snapshot(datetime.date(2025, 9, 8)),
            ]

            assert snapshot_list_fixture._filter_weekly(3, snap_list) == snap_list[:3]

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_non_sequential_weeks(self, snapshot_list_fixture):
            """snapshots on non-sequential weeks should be returned up to the requested number"""
            snap_list = [
                new_snapshot(datetime.date(2025, 9, 29)),
                new_snapshot(datetime.date(2025, 9, 15)),
                new_snapshot(datetime.date(2025, 9, 8)),
                new_snapshot(datetime.date(2025, 9, 1)),
            ]
            assert snapshot_list_fixture._filter_weekly(3, snap_list) == snap_list[:3]

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_max_one_per_week(self, snapshot_list_fixture):
            """only the most recent snapshot from any given calendar week should be considered"""
            snap_list = [
                new_snapshot(datetime.date(2025, 9, 30)),  # week 4
                new_snapshot(datetime.date(2025, 9, 28)),  # week 3
                new_snapshot(datetime.date(2025, 9, 25)),  # week 3
                new_snapshot(datetime.date(2025, 9, 22)),  # week 3
                new_snapshot(datetime.date(2025, 9, 21)),  # week 2
            ]
            assert snapshot_list_fixture._filter_weekly(3, snap_list) == [
                snap_list[0],
                snap_list[1],
                snap_list[4],
            ]

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_month_boundary(self, snapshot_list_fixture):
            """snapshots should be able to span the month boundary without issue"""
            snap_list = [
                new_snapshot(datetime.date(2025, 9, 23)),
                new_snapshot(datetime.date(2025, 9, 10)),
                new_snapshot(datetime.date(2025, 8, 23)),
                new_snapshot(datetime.date(2025, 8, 10)),
            ]
            assert snapshot_list_fixture._filter_weekly(3, snap_list) == snap_list[:3]

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_year_boundary(self, snapshot_list_fixture):
            """snapshots should be able to span the year boundary without issue"""
            snap_list = [
                new_snapshot(datetime.date(2025, 9, 23)),
                new_snapshot(datetime.date(2025, 9, 10)),
                new_snapshot(datetime.date(2024, 9, 23)),
                new_snapshot(datetime.date(2024, 9, 10)),
            ]
            assert snapshot_list_fixture._filter_weekly(3, snap_list) == snap_list[:3]

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_not_enough_snapshots(self, snapshot_list_fixture):
            """if more than the total number of weekly snapshots are requested, all weekly snapshots should be returned"""
            # snapshot_list_fixture has 5 snapshots from unique weeks, out of 10 total
            snap_list = snapshot_list_fixture.snapshots
            assert len(snapshot_list_fixture._filter_weekly(10, snap_list)) == 5

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_no_current(self, snapshot_list_fixture):
            """if no_current is set and a snapshot lies within the current time period, it should be excluded"""
            snap_list = [
                new_snapshot(datetime.date(2025, 9, 29)),
                new_snapshot(datetime.date(2025, 9, 22)),
                new_snapshot(datetime.date(2025, 9, 15)),
                new_snapshot(datetime.date(2025, 9, 8)),
            ]
            assert (
                snapshot_list_fixture._filter_weekly(
                    3,
                    snap_list,
                    no_current=True,
                    current_time=datetime.datetime(2025, 9, 30),
                )
                == snap_list[1:]
            )

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_snapshot_timezone_primacy(self, snapshot_list_fixture):
            """
            the timezone recorded in the snapshot should have primacy when determining in which period the snapshot falls
            e.g. 2018-11-03T01:38:41+01:00, and 2018-11-04T01:38:41+06:00, should be considered as having been taken on different days.
            this matches restic behaviour: https://forum.restic.net/t/which-timezone-does-forget-use/1166
            """
            snap_list = [
                new_snapshot(
                    datetime.datetime(
                        2018,
                        11,
                        5,
                        1,
                        38,
                        41,
                        tzinfo=datetime.timezone(datetime.timedelta(seconds=21600)),
                    )
                ),
                new_snapshot(
                    datetime.datetime(
                        2018,
                        11,
                        4,
                        1,
                        38,
                        41,
                        tzinfo=datetime.timezone(datetime.timedelta(seconds=3600)),
                    )
                ),
            ]
            assert (
                len(snapshot_list_fixture._filter_weekly(2, snap_list, no_current=True))
                == 2
            )

    class TestFilterMonthly:
        """Tests for the _filter_monthly method"""

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_zero(self, snapshot_list_fixture):
            """requesting zero snapshots should return an empty list"""
            snap_list = snapshot_list_fixture.snapshots
            assert snapshot_list_fixture._filter_monthly(0, snap_list) == []

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_sequential_months(self, snapshot_list_fixture):
            """snapshots on sequential months should be returned up to the requested number"""
            snap_list = [
                new_snapshot(datetime.date(2025, 9, 29)),
                new_snapshot(datetime.date(2025, 8, 11)),
                new_snapshot(datetime.date(2025, 7, 6)),
                new_snapshot(datetime.date(2025, 5, 31)),
            ]

            assert snapshot_list_fixture._filter_monthly(3, snap_list) == snap_list[:3]

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_non_sequential_months(self, snapshot_list_fixture):
            """snapshots on non-sequential months should be returned up to the requested number"""
            snap_list = [
                new_snapshot(datetime.date(2025, 9, 29)),
                new_snapshot(datetime.date(2025, 5, 31)),
                new_snapshot(datetime.date(2025, 4, 19)),
                new_snapshot(datetime.date(2025, 3, 1)),
            ]
            assert snapshot_list_fixture._filter_monthly(3, snap_list) == snap_list[:3]

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_max_one_per_month(self, snapshot_list_fixture):
            """only the most recent snapshot from any given calendar month should be considered"""
            snap_list = [
                new_snapshot(datetime.date(2025, 9, 30)),
                new_snapshot(datetime.date(2025, 9, 28)),
                new_snapshot(datetime.date(2025, 9, 25)),
                new_snapshot(datetime.date(2025, 8, 11)),
                new_snapshot(datetime.date(2025, 7, 6)),
            ]
            assert snapshot_list_fixture._filter_monthly(3, snap_list) == [
                snap_list[0],
                snap_list[3],
                snap_list[4],
            ]

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_year_boundary(self, snapshot_list_fixture):
            """snapshots should be able to span the year boundary without issue"""
            snap_list = [
                new_snapshot(datetime.date(2025, 9, 23)),
                new_snapshot(datetime.date(2025, 8, 11)),
                new_snapshot(datetime.date(2024, 9, 23)),
                new_snapshot(datetime.date(2024, 8, 11)),
            ]
            assert snapshot_list_fixture._filter_monthly(3, snap_list) == snap_list[:3]

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_not_enough_snapshots(self, snapshot_list_fixture):
            """if more than the total number of monthly snapshots are requested, all monthly snapshots should be returned"""
            # snapshot_list_fixture has 4 snapshots from unique months, out of 10 total
            snap_list = snapshot_list_fixture.snapshots
            assert len(snapshot_list_fixture._filter_monthly(10, snap_list)) == 4

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_no_current(self, snapshot_list_fixture):
            """if no_current is set and a snapshot lies within the current time period, it should be excluded"""
            snap_list = [
                new_snapshot(datetime.date(2025, 9, 29)),
                new_snapshot(datetime.date(2025, 8, 11)),
                new_snapshot(datetime.date(2025, 7, 6)),
                new_snapshot(datetime.date(2025, 5, 31)),
            ]
            assert (
                snapshot_list_fixture._filter_monthly(
                    3,
                    snap_list,
                    no_current=True,
                    current_time=datetime.datetime(2025, 9, 30),
                )
                == snap_list[1:]
            )

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_snapshot_timezone_primacy(self, snapshot_list_fixture):
            """
            the timezone recorded in the snapshot should have primacy when determining in which period the snapshot falls
            e.g. 2018-11-03T01:38:41+01:00, and 2018-11-04T01:38:41+06:00, should be considered as having been taken on different days.
            this matches restic behaviour: https://forum.restic.net/t/which-timezone-does-forget-use/1166
            """
            snap_list = [
                new_snapshot(
                    datetime.datetime(
                        2018,
                        11,
                        1,
                        1,
                        38,
                        41,
                        tzinfo=datetime.timezone(datetime.timedelta(seconds=21600)),
                    )
                ),
                new_snapshot(
                    datetime.datetime(
                        2018,
                        10,
                        31,
                        1,
                        38,
                        41,
                        tzinfo=datetime.timezone(datetime.timedelta(seconds=3600)),
                    )
                ),
            ]
            assert (
                len(
                    snapshot_list_fixture._filter_monthly(2, snap_list, no_current=True)
                )
                == 2
            )

    class TestFilterYearly:
        """Tests for the _filter_yearly method"""

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_zero(self, snapshot_list_fixture):
            """requesting zero snapshots should return an empty list"""
            snap_list = snapshot_list_fixture.snapshots
            assert snapshot_list_fixture._filter_yearly(0, snap_list) == []

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_sequential_years(self, snapshot_list_fixture):
            """snapshots on sequential years should be returned up to the requested number"""
            snap_list = [
                new_snapshot(datetime.date(2025, 9, 29)),
                new_snapshot(datetime.date(2024, 9, 23)),
                new_snapshot(datetime.date(2023, 7, 6)),
                new_snapshot(datetime.date(2022, 5, 31)),
            ]

            assert snapshot_list_fixture._filter_yearly(3, snap_list) == snap_list[:3]

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_non_sequential_years(self, snapshot_list_fixture):
            """snapshots on non-sequential years should be returned up to the requested number"""
            snap_list = [
                new_snapshot(datetime.date(2025, 9, 29)),
                new_snapshot(datetime.date(2022, 5, 31)),
                new_snapshot(datetime.date(2021, 3, 4)),
                new_snapshot(datetime.date(2020, 11, 15)),
            ]
            assert snapshot_list_fixture._filter_yearly(3, snap_list) == snap_list[:3]

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_max_one_per_year(self, snapshot_list_fixture):
            """only the most recent snapshot from any given calendar year should be considered"""
            snap_list = [
                new_snapshot(datetime.date(2025, 9, 30)),
                new_snapshot(datetime.date(2025, 8, 11)),
                new_snapshot(datetime.date(2025, 7, 6)),
                new_snapshot(datetime.date(2024, 9, 23)),
                new_snapshot(datetime.date(2023, 7, 6)),
            ]
            assert snapshot_list_fixture._filter_yearly(3, snap_list) == [
                snap_list[0],
                snap_list[3],
                snap_list[4],
            ]

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_not_enough_snapshots(self, snapshot_list_fixture):
            """if more than the total number of yearly snapshots are requested, all yearly snapshots should be returned"""
            # snapshot_list_fixture has 2 snapshots from unique years, out of 10 total
            snap_list = snapshot_list_fixture.snapshots
            assert len(snapshot_list_fixture._filter_yearly(10, snap_list)) == 2

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_no_current(self, snapshot_list_fixture):
            """if no_current is set and a snapshot lies within the current time period, it should be excluded"""
            snap_list = [
                new_snapshot(datetime.date(2025, 9, 29)),
                new_snapshot(datetime.date(2024, 9, 23)),
                new_snapshot(datetime.date(2023, 7, 6)),
                new_snapshot(datetime.date(2022, 5, 31)),
            ]
            assert (
                snapshot_list_fixture._filter_yearly(
                    3,
                    snap_list,
                    no_current=True,
                    current_time=datetime.datetime(2025, 11, 21),
                )
                == snap_list[1:]
            )

        @pytest.mark.usefixtures("snapshot_list_fixture")
        def test_snapshot_timezone_primacy(self, snapshot_list_fixture):
            """
            the timezone recorded in the snapshot should have primacy when determining in which period the snapshot falls
            e.g. 2018-11-03T01:38:41+01:00, and 2018-11-04T01:38:41+06:00, should be considered as having been taken on different days.
            this matches restic behaviour: https://forum.restic.net/t/which-timezone-does-forget-use/1166
            """
            snap_list = [
                new_snapshot(
                    datetime.datetime(
                        2019,
                        1,
                        1,
                        1,
                        38,
                        41,
                        tzinfo=datetime.timezone(datetime.timedelta(seconds=21600)),
                    )
                ),
                new_snapshot(
                    datetime.datetime(
                        2018,
                        12,
                        31,
                        1,
                        38,
                        41,
                        tzinfo=datetime.timezone(datetime.timedelta(seconds=3600)),
                    )
                ),
            ]
            assert (
                len(snapshot_list_fixture._filter_yearly(2, snap_list, no_current=True))
                == 2
            )



from importlib import metadata
import logging
from packaging import version
from pathlib import Path
import pytest
from subprocess import CompletedProcess
from unittest import mock

from restic_replica import __version__, __main__


def test_version():
    """The application version in pyproject.toml must match the application version in restic_replica/__init__.py"""
    assert version.parse(__version__) == version.parse(
        metadata.version("restic-replica")
    )


class TestMain:
    """Tests for the function __main__.main"""

    @pytest.fixture(autouse=True)
    def mock_ensure_config_file(self, monkeypatch):
        monkeypatch.setattr(
            "restic_replica.app.ensure_config_file", lambda *args, **kwargs: Path()
        )

    @pytest.fixture(autouse=True)
    def mock_read_config_file(self, monkeypatch):
        monkeypatch.setattr(
            "restic_replica.app.read_config_file",
            lambda *args, **kwargs: {
                "restic": {},
                "policy": {},
                "source": {},
                "destination": {},
            },
        )

    @pytest.fixture(autouse=True)
    def mock_get_logdir(self, monkeypatch):
        monkeypatch.setattr(
            "restic_replica.app.get_logdir", lambda *args, **kwargs: None
        )

    @pytest.fixture(autouse=True)
    def mock_setup_logging(self, monkeypatch, logger_fixture):
        monkeypatch.setattr(
            "restic_replica.console.setup_logging",
            lambda *args, **kwargs: logger_fixture,
        )

    @pytest.fixture(autouse=True)
    def mock_logging_headers(self, monkeypatch):
        monkeypatch.setattr(
            "restic_replica.console.logging_headers",
            lambda *args, **kwargs: None,
        )

    @pytest.fixture(autouse=True)
    def mock_get_restic(self, monkeypatch, restic_cli_fixture):
        monkeypatch.setattr(
            "restic_replica.app.get_restic", lambda *args, **kwargs: restic_cli_fixture
        )

    @pytest.fixture(autouse=True)
    def mock_get_policy(self, monkeypatch):
        monkeypatch.setattr(
            "restic_replica.app.get_policy", lambda *args, **kwargs: None
        )

    @pytest.fixture(autouse=True)
    def mock_get_repository(self, monkeypatch, repository_fixture):
        monkeypatch.setattr(
            "restic_replica.app.get_repository",
            lambda *args, **kwargs: repository_fixture,
        )

    @pytest.fixture(autouse=True)
    def mock_check_repository_access(self, monkeypatch):
        monkeypatch.setattr(
            "restic_replica.app.check_repository_access",
            lambda *args, **kwargs: True,
        )

    @pytest.fixture(autouse=True)
    def mock_copy_snapshots(self, monkeypatch):
        monkeypatch.setattr(
            "restic_replica.app.copy_snapshots",
            lambda *args, **kwargs: CompletedProcess(
                ["foo"], 0, stdout=b"did something\n"
            ),
        )

    def test_copy_success(self, caplog):
        """A successful copy should log an informational message"""
        caplog.set_level(logging.INFO)
        __main__.main([])
        assert caplog.records[2].message == "Finished copying snapshots"

    @mock.patch(
        "restic_replica.app.copy_snapshots",
        return_value=CompletedProcess(
            ["foo"], 0, stdout=None
        ),  # stdout will be None if no snapshots needed to be copied
    )
    def test_no_snapshots(self, mock_func_under_test, caplog):
        """If there are no snapshots to copy, an informational message should be logged"""
        caplog.set_level(logging.INFO)
        __main__.main([])
        assert (
            caplog.records[2].message
            == "All snapshots from the source are already present in the destination repository"
        )

    @mock.patch("restic_replica.app.check_repository_access", side_effect=RuntimeError)
    def test_repository_access_failure(self, mock_func_under_test):
        """A repository access failure should raise a SystemExit"""
        with pytest.raises(SystemExit):
            __main__.main([])
        # The destination repository should not be checked if the access check on the source fails
        mock_func_under_test.assert_called_once()

    @mock.patch("restic_replica.app.copy_snapshots", side_effect=RuntimeError)
    def test_copy_failure(self, *args):
        """A snapshot copy failure should raise a SystemExit"""
        with pytest.raises(SystemExit):
            __main__.main([])



from datetime import datetime
import pytest

from restic_replica.snapshots import SnapshotList
from tests.utils import new_snapshot


@pytest.fixture
def snapshot_list_fixture():
    return SnapshotList(
        [
            new_snapshot(datetime.fromisoformat("2023-12-31T15:19:14.968650111+01:00")),
            new_snapshot(datetime.fromisoformat("2025-04-04T15:19:14.968650111+01:00")),
            new_snapshot(datetime.fromisoformat("2025-08-31T15:19:14.968650111+01:00")),
            new_snapshot(datetime.fromisoformat("2025-09-10T15:19:14.968650111+01:00")),
            new_snapshot(datetime.fromisoformat("2025-09-14T15:19:14.968650111+01:00")),
            new_snapshot(datetime.fromisoformat("2025-09-15T15:19:14.968650111+01:00")),
            new_snapshot(datetime.fromisoformat("2025-09-20T15:19:14.968650111+01:00")),
            new_snapshot(datetime.fromisoformat("2025-09-21T07:01:42.000000000+01:00")),
            new_snapshot(datetime.fromisoformat("2025-09-21T10:34:09.000000000+01:00")),
            new_snapshot(datetime.fromisoformat("2025-09-21T15:19:14.968650111+01:00")),
        ]
    )



from contextlib import nullcontext as does_not_raise
import logging
import os
import pytest
import subprocess
from unittest import mock

from restic_replica import repository


class TestResticCli:
    """Tests for the class repository.ResticCli"""

    @pytest.fixture(autouse=True)
    def mock_os_environ(self, monkeypatch):
        monkeypatch.setattr(
            "os.environ",
            {},
        )

    class TestExecuteLiveOutput:
        """Tests for the _execute_live_output method"""

        @pytest.mark.usefixtures("restic_cli_fixture")
        def test_args(self, fp, restic_cli_fixture):
            """arguments should be passed through to the process"""
            args = ["restic", "snapshots"]
            fp.register(
                args,
                stdout=[
                    "repository a977efd9 opened (version 2, compression level auto)",
                    "ID        Time                 Host          Tags        Paths       Size",
                    "--------------------------------------------------------------------------",
                    "49833392  2025-09-08 15:41:12  oracca.local              /etc/hosts  384 B",
                    "--------------------------------------------------------------------------",
                    "1 snapshots",
                ],
            )
            process = restic_cli_fixture._execute_live_output(args)
            assert process.args == args
            assert process.returncode == 0

        @pytest.mark.usefixtures("restic_cli_fixture")
        def test_logger_output(self, fp, caplog, restic_cli_fixture):
            """output from the process should be emitted to logger line by line"""
            caplog.set_level(logging.INFO)
            args = ["restic", "snapshots"]
            fp.register(
                args,
                stdout=[
                    "repository a977efd9 opened (version 2, compression level auto)",
                    "Warning: at least one source file could not be read",
                    "Fatal: Please specify repository location (-r or --repository-file)",
                ],
            )

            restic_cli_fixture._execute_live_output(args)
            assert caplog.record_tuples == [
                (
                    "restic_replica.repository",
                    logging.INFO,
                    "repository a977efd9 opened (version 2, compression level auto)",
                ),
                (
                    "restic_replica.repository",
                    logging.WARNING,
                    "Warning: at least one source file could not be read",
                ),
                (
                    "restic_replica.repository",
                    logging.ERROR,
                    "Fatal: Please specify repository location (-r or --repository-file)",
                ),
            ]

        @pytest.mark.usefixtures("restic_cli_fixture")
        @pytest.mark.parametrize(
            "returncode, expectation",
            [
                (0, does_not_raise()),
                (1, pytest.raises(subprocess.CalledProcessError)),
                (2, pytest.raises(subprocess.CalledProcessError)),
                (3, does_not_raise()),
                (10, pytest.raises(subprocess.CalledProcessError)),
                (11, pytest.raises(subprocess.CalledProcessError)),
                (12, pytest.raises(subprocess.CalledProcessError)),
                (130, pytest.raises(subprocess.CalledProcessError)),
            ],
        )
        def test_returncodes(self, fp, restic_cli_fixture, returncode, expectation):
            """exit codes 0 and 3 should not raise an Exception, all other exit codes should"""
            args = ["restic", "snapshots"]
            with expectation:
                fp.register(args, returncode=returncode, stdout=None)
                process = restic_cli_fixture._execute_live_output(args)
                assert process.returncode == returncode

    class TestExecute:
        """Tests for the execute method"""

        @pytest.mark.usefixtures("restic_cli_fixture")
        def test_args_copy(self, fp, restic_cli_fixture):
            """input arguments should not be mutated"""
            args = ["snapshots"]
            args_with_restic = ["restic", "snapshots"]
            # register the args, prepending restic path
            fp.register(args_with_restic, stdout=None)
            # pass in the args without restic path prepended
            process = restic_cli_fixture.execute(args)
            # process args should have restic path prepended, original args list should should not be mutated
            assert args != args_with_restic
            assert process.args == args_with_restic

        @pytest.mark.usefixtures("restic_cli_fixture")
        def test_class_env_vars(self, restic_cli_fixture, monkeypatch):
            """
            environment variables should be set, with the ResticCli fixture environment
            variables having primacy. The input ductionary should not be mutated.
            """

            def return_environ(*args, **kwargs):
                """function that returns a copy of the os.environment dictionary"""
                return os.environ.copy()

            environment_vars = {
                "RESTIC_PROGRESS_FPS": "60",
                "RESTIC_PASSWORD": "secret",
            }
            monkeypatch.setattr("subprocess.run", return_environ)
            process_env = restic_cli_fixture.execute(["snapshots"], environment_vars)
            # variable already existing in class should be set to class value, not new value
            assert process_env["RESTIC_PROGRESS_FPS"] == "0.016667"
            # new variable should be added
            assert process_env["RESTIC_PASSWORD"] == "secret"
            # original environment_vars dict should not be mutated
            assert environment_vars == {
                "RESTIC_PROGRESS_FPS": "60",
                "RESTIC_PASSWORD": "secret",
            }

        @pytest.mark.usefixtures("restic_cli_fixture")
        def test_output_return(self, fp, restic_cli_fixture):
            """output from the command should be returned"""
            fake_stdout = [
                "repository a977efd9 opened (version 2, compression level auto)",
                "ID        Time                 Host          Tags        Paths       Size",
                "--------------------------------------------------------------------------",
                "49833392  2025-09-08 15:41:12  oracca.local              /etc/hosts  384 B",
                "--------------------------------------------------------------------------",
                "1 snapshots",
            ]
            # join the list with os.linesep(), and add trailing separator
            expected_stdout = (os.linesep).join(fake_stdout) + os.linesep
            fp.register(
                ["restic", "snapshots"],
                stdout=fake_stdout,
            )
            process = restic_cli_fixture.execute(["snapshots"])
            assert process.stdout == expected_stdout

        @pytest.mark.usefixtures("restic_cli_fixture")
        def test_json(self, fp, restic_cli_fixture):
            """the json parameter should be appended to the arguments"""
            fp.register(["restic", "snapshots", "--json"], stdout=None)
            process = restic_cli_fixture.execute(["snapshots"], json=True)
            assert process.args == ["restic", "snapshots", "--json"]

        @pytest.mark.usefixtures("restic_cli_fixture")
        def test_verbose(self, fp, restic_cli_fixture):
            """the verbose parameter should be appended to the arguments"""
            restic_cli_fixture.verbose = 1
            fp.register(["restic", "snapshots", "--verbose=1"], stdout=None)
            process = restic_cli_fixture.execute(["snapshots"])
            assert process.args == ["restic", "snapshots", "--verbose=1"]

        @pytest.mark.usefixtures("restic_cli_fixture")
        def test_live_output(self, restic_cli_fixture):
            """the live output process should return a populated CompletedProcess instance"""
            fake_return = subprocess.CompletedProcess(["./foo"], 0)
            with mock.patch.object(
                restic_cli_fixture, "_execute_live_output", return_value=fake_return
            ):
                assert (
                    restic_cli_fixture.execute(["snapshots"], live_output=True)
                    == fake_return
                )

        @pytest.mark.usefixtures("restic_cli_fixture")
        def test_environment_vars_unset(self, fp, restic_cli_fixture):
            """any environment variables that are set should be unset"""
            fp.register(["restic", "snapshots"])
            restic_cli_fixture.execute(["snapshots"])
            for key in restic_cli_fixture.environment_vars.keys():
                assert os.getenv(f"{key}") is None


class TestRepository:
    """Tests for the class repository.Repository"""

    class TestInit:
        """Tests for the __init__ method"""

        @pytest.mark.usefixtures("repository_fixture")
        def test_environment_vars_merge(self, repository_fixture):
            """
            supplied password should result in the correct environment variable being
            appended to the existing environment variables dictionary.
            """
            assert repository_fixture.environment_vars == {
                "RESTIC_COMPRESSION": "max",
                "RESTIC_PASSWORD": "secret",
            }

        @pytest.mark.usefixtures("restic_cli_fixture")
        def test_password_file_environment_var(self, restic_cli_fixture):
            """supplied password file should result in the correct environment variable being set"""
            myrepo = repository.Repository(
                "/tmp/myrepo",
                "myrepo",
                restic_cli_fixture,
                password_file="/path/to/password-file",
            )
            assert myrepo.environment_vars == {
                "RESTIC_PASSWORD_FILE": "/path/to/password-file"
            }

        @pytest.mark.usefixtures("restic_cli_fixture")
        def test_password_command_environment_var(self, restic_cli_fixture):
            """supplied password command should result in the correct environment variable being set"""
            myrepo = repository.Repository(
                "/tmp/myrepo",
                "myrepo",
                restic_cli_fixture,
                password_command="/path/to/password-command",
            )
            assert myrepo.environment_vars == {
                "RESTIC_PASSWORD_COMMAND": "/path/to/password-command"
            }

    class TestVerifyPasswordIsSet:
        """Tests for the _verify_password_is_set method"""

        @pytest.mark.usefixtures("repository_fixture")
        def test_password(self, repository_fixture):
            """A set password should return true"""
            assert repository_fixture._verify_password_is_set() is True

        @pytest.mark.usefixtures("repository_fixture")
        def test_password_file(self, repository_fixture):
            """A set password_file should return true"""
            repository_fixture.password_file = "/path/to/password-file"
            repository_fixture.password = None
            assert repository_fixture._verify_password_is_set() is True

        @pytest.mark.usefixtures("repository_fixture")
        def test_password_command(self, repository_fixture):
            """A set password_command should return true"""
            repository_fixture.password_command = "/path/to/password-command"
            repository_fixture.password = None
            assert repository_fixture._verify_password_is_set() is True

        @pytest.mark.usefixtures("repository_fixture")
        def test_password_file_warning(self, repository_fixture, caplog):
            """A set password and password_file should return True but emit a warning to logger"""
            repository_fixture.password_file = "/path/to/password-file"
            assert repository_fixture._verify_password_is_set() is True
            assert caplog.record_tuples[0] == (
                "restic_replica.repository",
                30,
                f"password and password_file were specified for repository {repository_fixture.uri}; password_file will be used",
            )

        @pytest.mark.usefixtures("repository_fixture")
        def test_password_command_warning(self, repository_fixture, caplog):
            """A set password and password_command should return True but emit a warning to logger"""
            repository_fixture.password_command = "/path/to/password-command"
            assert repository_fixture._verify_password_is_set() is True
            assert caplog.record_tuples[0] == (
                "restic_replica.repository",
                30,
                f"password and password_command were specified for repository {repository_fixture.uri}; password_command will be used",
            )

        @pytest.mark.usefixtures("repository_fixture")
        def test_password_file_and_command(self, repository_fixture):
            """A set password_file and password_command should raise a KeyError exception"""
            repository_fixture.password_file = "/path/to/password-file"
            repository_fixture.password_command = "/path/to/password-command"
            repository_fixture.password = None
            with pytest.raises(KeyError):
                repository_fixture._verify_password_is_set()

        @pytest.mark.usefixtures("repository_fixture")
        def test_no_password_attributes(self, repository_fixture):
            """If no password* attribute is set, a ValueError exception should be raised"""
            repository_fixture.password = None
            with pytest.raises(ValueError):
                repository_fixture._verify_password_is_set()

    class TestFilterOtherEnv:
        """Tests for the method _filter_other_env"""

        def test_filter(self, repository_fixture):
            """RESTIC_PASSWORD, RESTIC_PASSWORD_FILE, and RESTIC_PASSWORD_COMMAND should be filtered out"""
            environment_vars = {
                "RESTIC_COMPRESSION": "max",
                "RESTIC_PASSWORD": "secret",
                "RESTIC_PASSWORD_FILE": "/path/to/password-file",
                "RESTIC_PASSWORD_COMMAND": "/path/to/password-command",
            }
            assert dict(
                filter(repository_fixture._filter_other_env, environment_vars.items())
            ) == {"RESTIC_COMPRESSION": "max"}

    class TestSnapshots:
        """Tests for the snapshots method"""

        def return_args(self, *args, **kwargs):
            """function that returns all args passed to it"""
            return args

        def return_kwargs(self, *args, **kwargs):
            """function that returns all kwargs passed to it"""
            return kwargs

        @pytest.mark.usefixtures("repository_fixture")
        def test_args(self, repository_fixture):
            """args should include _common_args and have `snapshots` appended"""
            with mock.patch.object(
                repository_fixture.restic_cli,
                "execute",
                self.return_args,
            ):
                assert repository_fixture.snapshots() == (
                    [
                        "-r",
                        f"{repository_fixture.uri}",
                        "snapshots",
                    ],
                )

        @pytest.mark.usefixtures("repository_fixture")
        def test_environment_vars(self, repository_fixture):
            """environment vars should be passed as-is to restic_cli"""
            with mock.patch.object(
                repository_fixture.restic_cli,
                "execute",
                self.return_kwargs,
            ):
                assert (
                    repository_fixture.snapshots()["environment_vars"]
                    == repository_fixture.environment_vars
                )

        @pytest.mark.usefixtures("repository_fixture")
        def test_live_output(self, repository_fixture):
            """live_output var should be set to false by default"""
            with mock.patch.object(
                repository_fixture.restic_cli,
                "execute",
                self.return_kwargs,
            ):
                assert repository_fixture.snapshots()["live_output"] is False
                assert (
                    repository_fixture.snapshots(live_output=True)["live_output"]
                    is True
                )

        @pytest.mark.usefixtures("repository_fixture")
        def test_json(self, repository_fixture):
            """json var should be set to false by default"""
            with mock.patch.object(
                repository_fixture.restic_cli,
                "execute",
                self.return_kwargs,
            ):
                assert repository_fixture.snapshots()["json"] is False
                assert repository_fixture.snapshots(json=True)["json"] is True

    class TestCopy:
        """Tests for the copy method"""

        def return_args(self, *args, **kwargs):
            """function that returns all args passed to it"""
            return args

        def return_kwargs(self, *args, **kwargs):
            """function that returns all kwargs passed to it"""
            return kwargs

        @pytest.fixture
        def other_repository_fixture(self, restic_cli_fixture):
            """Return a (:class:`repository.Repository`) instance"""
            return repository.Repository(
                "/tmp/repo2",
                "repo2",
                restic_cli_fixture,
                "secret2",
                environment_vars={"RESTIC_COMPRESSION": "off", "RESTIC_PACK_SIZE": 128},
            )

        @pytest.mark.usefixtures("repository_fixture")
        def test_same_repository(self, repository_fixture):
            """if the same repository is set as both source and destination a RuntimeError should be raised"""
            with pytest.raises(RuntimeError):
                repository_fixture.copy(repository_fixture)

        @pytest.mark.usefixtures("repository_fixture", "other_repository_fixture")
        def test_args(self, repository_fixture, other_repository_fixture):
            """args should include _common_args and have `copy` and second repository appended. No snapshot IDs should be supplied"""
            with mock.patch.object(
                repository_fixture.restic_cli,
                "execute",
                self.return_args,
            ):
                assert repository_fixture.copy(other_repository_fixture) == (
                    [
                        "-r",
                        f"{repository_fixture.uri}",
                        "copy",
                        "--from-repo",
                        f"{other_repository_fixture.uri}",
                    ],
                )

        @pytest.mark.usefixtures("repository_fixture", "other_repository_fixture")
        def test_other_password(self, repository_fixture, other_repository_fixture):
            """environment vars should be passed as-is to restic_cli, but RESTIC_FROM_PASSWORD should also be set"""
            with mock.patch.object(
                repository_fixture.restic_cli,
                "execute",
                self.return_kwargs,
            ):
                expected_environment_vars = (
                    other_repository_fixture.environment_vars.copy()
                )
                expected_environment_vars.update(repository_fixture.environment_vars)
                expected_environment_vars["RESTIC_FROM_PASSWORD"] = (
                    other_repository_fixture.password
                )
                assert (
                    repository_fixture.copy(other_repository_fixture)[
                        "environment_vars"
                    ]
                    == expected_environment_vars
                )

        @pytest.mark.usefixtures("repository_fixture", "other_repository_fixture")
        def test_other_password_file(
            self, repository_fixture, other_repository_fixture
        ):
            """environment vars should be passed as-is to restic_cli, but RESTIC_FROM_PASSWORD_FILE should also be set"""
            with mock.patch.object(
                repository_fixture.restic_cli,
                "execute",
                self.return_kwargs,
            ):
                # switch password for password_file
                other_repository_fixture.password_file = "/path/to/password-file2"
                repository_fixture.password_file = "/path/to/password-file"
                other_repository_fixture.environment_vars.pop("RESTIC_PASSWORD", None)
                repository_fixture.environment_vars.pop("RESTIC_PASSWORD", None)

                expected_environment_vars = (
                    other_repository_fixture.environment_vars.copy()
                )
                expected_environment_vars.update(repository_fixture.environment_vars)
                expected_environment_vars["RESTIC_FROM_PASSWORD_FILE"] = (
                    other_repository_fixture.password_file
                )
                assert (
                    repository_fixture.copy(other_repository_fixture)[
                        "environment_vars"
                    ]
                    == expected_environment_vars
                )

        @pytest.mark.usefixtures("repository_fixture", "other_repository_fixture")
        def test_other_password_command(
            self, repository_fixture, other_repository_fixture
        ):
            """environment vars should be passed as-is to restic_cli, but RESTIC_FROM_PASSWORD_COMMAND should also be set"""
            with mock.patch.object(
                repository_fixture.restic_cli,
                "execute",
                self.return_kwargs,
            ):
                # switch password for password_command
                other_repository_fixture.password_command = "/path/to/password-command2"
                repository_fixture.password_command = "/path/to/password-command"
                other_repository_fixture.environment_vars.pop("RESTIC_PASSWORD", None)
                repository_fixture.environment_vars.pop("RESTIC_PASSWORD", None)

                expected_environment_vars = (
                    other_repository_fixture.environment_vars.copy()
                )
                expected_environment_vars.update(repository_fixture.environment_vars)
                expected_environment_vars["RESTIC_FROM_PASSWORD_COMMAND"] = (
                    other_repository_fixture.password_command
                )
                assert (
                    repository_fixture.copy(other_repository_fixture)[
                        "environment_vars"
                    ]
                    == expected_environment_vars
                )

        @pytest.mark.usefixtures("repository_fixture", "other_repository_fixture")
        def test_other_environment_vars(
            self, repository_fixture, other_repository_fixture
        ):
            """the other repository environment vars should also be set, unless they conflict with the current repository environment vars"""
            with mock.patch.object(
                repository_fixture.restic_cli,
                "execute",
                self.return_kwargs,
            ):
                other_repository_fixture.password_command = "/path/to/password-command2"
                expected_environment_vars = {}
                expected_environment_vars.update(
                    other_repository_fixture.environment_vars
                )
                expected_environment_vars.pop("RESTIC_PASSWORD", None)
                expected_environment_vars.pop("RESTIC_PASSWORD_FILE", None)
                expected_environment_vars.pop("RESTIC_PASSWORD_COMMAND", None)
                expected_environment_vars.update(repository_fixture.environment_vars)
                expected_environment_vars["RESTIC_FROM_PASSWORD_COMMAND"] = (
                    other_repository_fixture.password_command
                )
                assert (
                    repository_fixture.copy(other_repository_fixture)[
                        "environment_vars"
                    ]
                    == expected_environment_vars
                )

        @pytest.mark.usefixtures("repository_fixture", "other_repository_fixture")
        def test_live_output(self, repository_fixture, other_repository_fixture):
            """live_output should be set to false by default"""
            with mock.patch.object(
                repository_fixture.restic_cli,
                "execute",
                self.return_kwargs,
            ):
                assert (
                    repository_fixture.copy(other_repository_fixture)["live_output"]
                    is False
                )
                assert (
                    repository_fixture.copy(other_repository_fixture, live_output=True)[
                        "live_output"
                    ]
                    is True
                )

        @pytest.mark.usefixtures("repository_fixture", "other_repository_fixture")
        def test_json(self, repository_fixture, other_repository_fixture):
            """json should be set to false by default"""
            with mock.patch.object(
                repository_fixture.restic_cli,
                "execute",
                self.return_kwargs,
            ):
                assert (
                    repository_fixture.copy(other_repository_fixture)["json"] is False
                )
                assert (
                    repository_fixture.copy(other_repository_fixture, json=True)["json"]
                    is True
                )

        @pytest.mark.usefixtures(
            "repository_fixture", "other_repository_fixture", "snapshot_list_fixture"
        )
        def test_copy_snapshotlist(
            self, repository_fixture, other_repository_fixture, snapshot_list_fixture
        ):
            """supplying a snapshot list should result in all snapshot IDs being appended to args"""

            expected_ids = [i.id for i in snapshot_list_fixture.snapshots]
            with mock.patch.object(
                repository_fixture.restic_cli,
                "execute",
                self.return_args,
            ):
                assert repository_fixture.copy(
                    other_repository_fixture, snapshots=snapshot_list_fixture
                ) == (
                    ["-r", "/tmp/restic-repo", "copy", "--from-repo", "/tmp/repo2"]
                    + expected_ids,
                )



import pytest

from restic_replica.repository import Repository


@pytest.fixture
def repository_fixture(restic_cli_fixture):
    """Return a (:class:`repository.Repository`) instance"""
    return Repository(
        "/tmp/restic-repo",
        "myrepo",
        restic_cli_fixture,
        password="secret",
        environment_vars={"RESTIC_COMPRESSION": "max"},
    )



import importlib.resources
from pathlib import Path
import pytest
from subprocess import CalledProcessError, CompletedProcess
import textwrap
import tomllib
from unittest import mock

from restic_replica import __assets__, app
from restic_replica.repository import Repository, ResticCli
from restic_replica.snapshots import Policy


class TestEnsureConfigFile:
    """Tests for the function app.ensure_config_file"""

    def test_existing_config_file(self, tmp_path):
        """
        if a path to a config file is supplied, and the file exists, the path to the
        file should be returned.
        """
        assert app.ensure_config_file(tmp_path) == tmp_path

    @mock.patch("pathlib.Path.exists", return_value=True)
    @mock.patch("platform.system", return_value="Linux")
    def test_existing_default_config_file_nonwin(self, *args):
        """
        if no path to a config file is supplied, and the platform is non-windows, and
        the default config file exists, the default config file path for non-windows
        should be returned.
        """
        assert app.ensure_config_file() == Path.home() / ".restic-replica/config.toml"

    @mock.patch("pathlib.Path.exists", return_value=True)
    @mock.patch("platform.system", return_value="Windows")
    def test_existing_default_config_file_win(self, *args):
        """
        if no path to a config file is supplied, and the platform is windows, and the
        default config file exists, the default config file path for windows should be
        returned.
        """
        assert (
            app.ensure_config_file()
            == Path.home() / "AppData/Local/restic-replica/config.toml"
        )

    @mock.patch("pathlib.Path.exists", return_value=False)
    @mock.patch("platform.system", return_value="Linux")
    @mock.patch("pathlib.Path.mkdir", return_value=None)
    @mock.patch("shutil.copyfile", return_value=None)
    def test_default_config_file_path_nonwin(self, *args):
        """
        if no path to a config file is supplied, and the platform is non-windows, and
        the default config file does not exist, a SystemExit exception should be raised.
        """
        with pytest.raises(SystemExit):
            app.ensure_config_file()
        args[0].assert_called_with(
            importlib.resources.files(__assets__) / "example_config.toml",
            Path.home() / ".restic-replica/config.toml",
        )

    @mock.patch("pathlib.Path.exists", return_value=False)
    @mock.patch("platform.system", return_value="Windows")
    @mock.patch("pathlib.Path.mkdir", return_value=None)
    @mock.patch("shutil.copyfile", return_value=None)
    def test_default_config_file_path_win(self, *args):
        """
        if no path to a config file is supplied, and the platform is windows, and the
        default config file does not exist, a SystemExit exception should be raised.
        """
        with pytest.raises(SystemExit):
            app.ensure_config_file()
        args[0].assert_called_with(
            importlib.resources.files(__assets__) / "example_config_win.toml",
            Path.home() / "AppData/Local/restic-replica/config.toml",
        )

    @mock.patch("platform.system", return_value="Linux")
    @mock.patch("pathlib.Path.mkdir", return_value=None)
    @mock.patch("shutil.copyfile", return_value=None)
    def test_missing_config_file(self, *args):
        """
        if a path to a config file is supplied, and the file does not exist, a
        SystemExit exception should be raised.
        """
        target = Path("/not/a/real/path")
        with pytest.raises(SystemExit):
            app.ensure_config_file(target)
        args[0].assert_called_with(
            importlib.resources.files(__assets__) / "example_config.toml",
            target,
        )


class TestReadConfigFile:
    """Tests for the function app.read_config_file"""

    def test_valid_toml(self, tmp_path):
        """Valid toml should load into a dictionary correctly"""
        good_toml = textwrap.dedent(
            """\
            [app]
            restic_binary = "/usr/local/bin/restic"
            
            [source]
            repository_uri = "/tmp/restic-repo"
            password = "secret"
        """
        )
        expected_result = {
            "app": {"restic_binary": "/usr/local/bin/restic"},
            "source": {"repository_uri": "/tmp/restic-repo", "password": "secret"},
        }
        f = tmp_path / "good.toml"
        f.write_text(good_toml, encoding="utf-8")
        assert app.read_config_file(f) == expected_result

    def test_invalid_toml(self, tmp_path):
        """Invalid toml should raise TOMLDecodeError"""
        bad_toml = textwrap.dedent(
            """\
            [app]
            restic_binary = "/usr/local/bin/restic
            
            [source
            repository_uri = "/tmp/restic-repo"
            password = "secret"
        """
        )
        f = tmp_path / "bad.toml"
        f.write_text(bad_toml, encoding="utf-8")
        with pytest.raises(tomllib.TOMLDecodeError):
            app.read_config_file(f)

    def test_missing_file(self, tmp_path):
        """An invalid/nonexistent file path should raise a FileNotFoundError"""
        f = tmp_path / "notarealfile.toml"
        with pytest.raises(FileNotFoundError):
            app.read_config_file(f)


class TestGetLogdir:
    """Tests for the function app.get_logdir"""

    def test_provided_logdir(self):
        """A log directory path should be returned if one is provided"""
        assert app.get_logdir(
            {"app": {"log_directory": "/var/log/restic-replica/"}}
        ) == Path("/var/log/restic-replica/")

    def test_provided_logdir_tilde(self):
        """If tilde is used in the directory path, it should be expanded"""
        assert app.get_logdir({"app": {"log_directory": "~/.restic-replica/"}}) == Path(
            Path.home() / ".restic-replica/"
        )

    def test_missing_logdir_nonwin(self):
        """The default log directory for not-windows should be returned if no log directory is provided"""
        with mock.patch("platform.system", return_value="Linux"):
            assert app.get_logdir({}) == Path.home() / ".restic-replica"

    def test_missing_logdir_windows(self):
        """The default log directory for windows should be returned if no log directory is provided"""
        with mock.patch("platform.system", return_value="Windows"):
            assert app.get_logdir({}) == Path.home() / "AppData/Local/restic-replica"


class TestGetRestic:
    """Tests for the function app.get_restic"""

    def test_defaults_nonwin(self):
        """Default configuration for not-windows should be set if no arguments are supplied"""
        with mock.patch("platform.system", return_value="Linux"):
            assert app.get_restic({}) == ResticCli(
                Path("restic"), {"RESTIC_PROGRESS_FPS": "0.016667"}
            )

    def test_defaults_win(self):
        """Default configuration for windows should be set if no arguments are supplied"""
        with mock.patch("platform.system", return_value="Windows"):
            assert app.get_restic({}) == ResticCli(
                Path("restic.exe"), {"RESTIC_PROGRESS_FPS": "0.016667"}
            )

    def test_verbose(self):
        """A supplied verbosity level should be should be included in the returned class instance"""
        assert app.get_restic({}, verbose=2).verbose == 2

    def test_path(self):
        """A supplied restic path should be included in the returned class instance"""
        config = {"path": "/usr/local/bin/restic"}
        assert app.get_restic(config).path == Path("/usr/local/bin/restic")

    def test_progress_fps(self):
        """A supplied value for RESTIC_PROGRESS_FPS should be included in the returned class instance"""
        config = {"environment": {"RESTIC_PROGRESS_FPS": "0.003333"}}
        assert app.get_restic(config).environment_vars == {
            "RESTIC_PROGRESS_FPS": "0.003333"
        }


class TestGetPolicy:
    """Tests for the function app.get_policy"""

    def test_no_policy(self):
        """an empty dictionary (i.e. the user set no policy options) should return None"""
        assert app.get_policy({}) is None

    @pytest.mark.parametrize(
        "policy, expectation",
        [
            ({"keep-last": 10}, Policy(10, 0, 0, 0, 0)),
            ({"keep-daily": 10}, Policy(0, 10, 0, 0, 0)),
            ({"keep-weekly": 10}, Policy(0, 0, 10, 0, 0)),
            ({"keep-monthly": 10}, Policy(0, 0, 0, 10, 0)),
            ({"keep-yearly": 10}, Policy(0, 0, 0, 0, 10)),
        ],
    )
    def test_policy_options(self, policy, expectation):
        """one or more policy options should return a Policy instance, with any unset options set to 0 (disabled)"""
        assert app.get_policy(policy) == expectation

    def test_all_policy_options(self):
        """all policy options should be able to be set at once"""
        assert app.get_policy(
            {
                "keep-last": 10,
                "keep-daily": 9,
                "keep-weekly": 8,
                "keep-monthly": 7,
                "keep-yearly": 6,
            }
        ) == Policy(10, 9, 8, 7, 6)

    def test_non_integer_input(self):
        """an invalid policy input (i.e. not-an-integer, or a negative integer) should raise a RuntimeError"""
        with pytest.raises(RuntimeError):
            app.get_policy({"keep-last": "foo"})

    def test_invalid_policy(self):
        """inputs that would lead to an invalid policy (i.e. all zeroes) should raise a RuntimeError"""
        with pytest.raises(RuntimeError):
            app.get_policy({"keep-last": 0})

    @pytest.mark.parametrize(
        "state",
        [True, False],
    )
    def test_no_current_set(self, state):
        """setting exclude-current-period should set no_current in the returned Policy instance"""
        assert app.get_policy(
            {"keep-last": 1, "exclude-current-period": state}
        ) == Policy(1, no_current=state)

    def test_no_current_unset(self):
        """omitting exclude-current-period should set no_current to False in the returned Policy instance"""
        assert app.get_policy({"keep-last": 1}) == Policy(1, no_current=False)


class TestGetRepository:
    """Tests for the function app.get_repository"""

    @pytest.mark.usefixtures("restic_cli_fixture")
    def test_password(self, restic_cli_fixture):
        """A supplied password should be included in the instanced repository"""
        config = {"repository_uri": "/tmp/restic-repo", "password": "secret"}
        assert app.get_repository("myrepo", config, restic_cli_fixture) == Repository(
            "/tmp/restic-repo", "myrepo", restic_cli_fixture, password="secret"
        )

    @pytest.mark.usefixtures("restic_cli_fixture")
    def test_password_file(self, restic_cli_fixture):
        """A supplied password_file should be included in the instanced repository"""
        config = {
            "repository_uri": "/tmp/restic-repo",
            "password_file": "/path/to/secret",
        }
        assert app.get_repository("myrepo", config, restic_cli_fixture) == Repository(
            "/tmp/restic-repo",
            "myrepo",
            restic_cli_fixture,
            password_file="/path/to/secret",
        )

    @pytest.mark.usefixtures("restic_cli_fixture")
    def test_password_command(self, restic_cli_fixture):
        """A supplied password_command should be included in the instanced repository"""
        config = {
            "repository_uri": "/tmp/restic-repo",
            "password_command": "/bin/getsecret myrepo",
        }
        assert app.get_repository("myrepo", config, restic_cli_fixture) == Repository(
            "/tmp/restic-repo",
            "myrepo",
            restic_cli_fixture,
            password_command="/bin/getsecret myrepo",
        )

    @pytest.mark.usefixtures("restic_cli_fixture")
    def test_environment_variables(self, restic_cli_fixture):
        """Any supplied environment variables should be included in the instanced repository"""
        config = {
            "repository_uri": "/tmp/restic-repo",
            "password": "secret",
            "environment": {"RESTIC_COMPRESSION": "true"},
        }
        assert app.get_repository("myrepo", config, restic_cli_fixture) == Repository(
            "/tmp/restic-repo",
            "myrepo",
            restic_cli_fixture,
            password="secret",
            environment_vars={"RESTIC_COMPRESSION": "true"},
        )


class TestCheckRepositoryAccess:
    """Tests for the function app.check_repository_access"""

    @pytest.mark.usefixtures("repository_fixture")
    def test_valid_repository(self, repository_fixture):
        """Should return True in the event of successful access check"""
        with mock.patch.object(
            repository_fixture, "snapshots", return_value=CompletedProcess(["./foo"], 0)
        ):
            assert app.check_repository_access(repository_fixture)

    @pytest.mark.usefixtures("repository_fixture")
    def test_invalid_repository(self, repository_fixture):
        """Should raise a RuntimeError if the repository is unable to be accessed"""
        with mock.patch.object(
            repository_fixture,
            "snapshots",
            side_effect=CalledProcessError(1, "notalrealcommand"),
        ):
            with pytest.raises(RuntimeError):
                app.check_repository_access(repository_fixture)

    @pytest.mark.usefixtures("repository_fixture")
    def test_restic_error(self, repository_fixture):
        """Should raise a RuntimeError if the OS throws an exception accessing restic"""
        with mock.patch.object(
            repository_fixture,
            "snapshots",
            side_effect=FileNotFoundError(
                "[Errno 2] No such file or directory: '/usr/local/bin/restic'"
            ),
        ):
            with pytest.raises(RuntimeError):
                app.check_repository_access(repository_fixture)


class TestGetFilteredSnapshots:
    """Tests for the function app.get_filtered_snapshots"""

    def test_some_snapshots(
        self, repository_fixture, snapshot_list_fixture, snapshot_fixture, monkeypatch
    ):
        """if filter returns some snapshots a SnapshotList instance should be returned"""
        monkeypatch.setattr(
            repository_fixture,
            "snapshots",
            lambda *args, **kwargs: CompletedProcess(["./foo"], 0, None),
        )
        monkeypatch.setattr(
            "restic_replica.snapshots.SnapshotList.from_json",
            lambda *args, **kwargs: snapshot_list_fixture,
        )
        monkeypatch.setattr(
            "restic_replica.snapshots.SnapshotList.filter",
            lambda *args, **kwargs: [snapshot_fixture],
        )
        assert app.get_filtered_snapshots(repository_fixture, Policy(1)).snapshots == [
            snapshot_fixture
        ]

    def test_no_snapshots(
        self, repository_fixture, snapshot_list_fixture, snapshot_fixture, monkeypatch
    ):
        """if filter returns no snapshots a RuntimeError should be raised"""
        monkeypatch.setattr(
            repository_fixture,
            "snapshots",
            lambda *args, **kwargs: CompletedProcess(["./foo"], 0, None),
        )
        monkeypatch.setattr(
            "restic_replica.snapshots.SnapshotList.from_json",
            lambda *args, **kwargs: snapshot_list_fixture,
        )
        monkeypatch.setattr(
            "restic_replica.snapshots.SnapshotList.filter",
            lambda *args, **kwargs: [],
        )
        with pytest.raises(RuntimeError):
            app.get_filtered_snapshots(repository_fixture, Policy(1)).snapshots


class TestCopySnapshots:
    """Tests for the function app.copy_snapshots"""

    def return_kwargs(self, *args, **kwargs):
        """function that returns all kwargs passed to it"""
        return kwargs

    @pytest.mark.usefixtures("repository_fixture", "restic_cli_fixture")
    def test_copy_success(self, repository_fixture, restic_cli_fixture):
        """Should return true if the copy operation is successful"""
        with mock.patch.object(
            repository_fixture, "copy", return_value=CompletedProcess(["./foo"], 0)
        ):
            assert isinstance(
                app.copy_snapshots(
                    Repository(
                        "/tmp/restic-repo2",
                        "myrepo2",
                        restic_cli_fixture,
                        password="secret2",
                    ),
                    repository_fixture,
                ),
                CompletedProcess,
            )

    @pytest.mark.usefixtures("repository_fixture", "restic_cli_fixture")
    def test_copy_fail(self, repository_fixture, restic_cli_fixture):
        """Should raise RuntimeError if the copy operation fails"""
        with mock.patch.object(
            repository_fixture,
            "copy",
            side_effect=CalledProcessError(1, "notalrealcommand"),
        ):
            with pytest.raises(RuntimeError):
                app.copy_snapshots(
                    repository_fixture,
                    Repository(
                        "/tmp/restic-repo2",
                        "myrepo2",
                        restic_cli_fixture,
                        password="secret2",
                    ),
                )

    @pytest.mark.usefixtures("repository_fixture", "restic_cli_fixture")
    def test_restic_error(self, repository_fixture, restic_cli_fixture):
        """Should raise a RuntimeError if the OS throws an exception accessing restic"""
        with mock.patch.object(
            repository_fixture,
            "copy",
            side_effect=FileNotFoundError(
                "[Errno 2] No such file or directory: '/usr/local/bin/restic'"
            ),
        ):
            with pytest.raises(RuntimeError):
                app.copy_snapshots(
                    repository_fixture,
                    Repository(
                        "/tmp/restic-repo2",
                        "myrepo2",
                        restic_cli_fixture,
                        password="secret2",
                    ),
                )

    @pytest.mark.usefixtures(
        "repository_fixture", "restic_cli_fixture", "snapshot_list_fixture"
    )
    def test_snapshot_list(
        self, repository_fixture, restic_cli_fixture, snapshot_list_fixture
    ):
        """A SnapshotList object should be passed if a policy is provided"""
        with mock.patch.object(repository_fixture, "copy", self.return_kwargs):
            with mock.patch(
                "restic_replica.app.get_filtered_snapshots",
                return_value=snapshot_list_fixture,
            ):
                result = app.copy_snapshots(
                    Repository(
                        "/tmp/restic-repo2",
                        "myrepo2",
                        restic_cli_fixture,
                        password="secret2",
                    ),
                    repository_fixture,
                    policy=Policy(1),
                )
                assert result["snapshots"] == snapshot_list_fixture

    @pytest.mark.usefixtures("repository_fixture", "restic_cli_fixture")
    def test_no_snapshot_list(self, repository_fixture, restic_cli_fixture):
        """A SnapshotList object should not be passed if a policy is not provided"""
        with mock.patch.object(repository_fixture, "copy", self.return_kwargs):
            result = app.copy_snapshots(
                Repository(
                    "/tmp/restic-repo2",
                    "myrepo2",
                    restic_cli_fixture,
                    password="secret2",
                ),
                repository_fixture,
            )
            assert result["snapshots"] is None

    @pytest.mark.usefixtures("repository_fixture", "restic_cli_fixture")
    def test_dry_run(self, repository_fixture, restic_cli_fixture):
        """if the dry_run argument is passed, systemexit should be raised, and copy should not be called"""
        with mock.patch.object(repository_fixture, "copy", side_effect=RuntimeError):
            with pytest.raises(SystemExit):
                app.copy_snapshots(
                    Repository(
                        "/tmp/restic-repo2",
                        "myrepo2",
                        restic_cli_fixture,
                        password="secret2",
                    ),
                    repository_fixture,
                    dry_run=True,
                )



import importlib.resources
import logging
from pathlib import Path
import platform
import shutil
from subprocess import CalledProcessError, CompletedProcess
import tomllib
from typing import Optional

from restic_replica import __assets__
from restic_replica.repository import Repository, ResticCli
from restic_replica.snapshots import Policy, SnapshotList

logger = logging.getLogger(__name__)


def ensure_config_file(config_file: Optional[Path] = None) -> Path:
    """
    Search for config file in expected location. If one does not exist, create one, then raise SystemExit.

    Note: This function is called pre-logging setup, so any messages are printed to
    stdout.

    Args:
        config_file: path to the application configuration file, which may or may not exist.

    Returns:
        config_file: path to the application configuration file, which is confirmed to exist.

    Raises:
        SystemExit: if no configuration file exists, this exception is raised after an example configuration file has been created.
    """
    # set default path if one was not supplied
    if not config_file:
        if platform.system() == "Windows":
            config_file = Path.home() / "AppData/Local/restic-replica/config.toml"
        else:
            config_file = Path.home() / ".restic-replica" / "config.toml"
    # create config file and parent dir if config file does not exist
    if not config_file.exists():
        print("ERROR: Missing configuration file")
        config_file.parent.mkdir(parents=True, exist_ok=True)
        if platform.system() == "Windows":
            shutil.copyfile(
                importlib.resources.files(__assets__) / "example_config_win.toml",
                config_file,
            )
        else:
            shutil.copyfile(
                importlib.resources.files(__assets__) / "example_config.toml",
                config_file,
            )
        print(
            f"An example configuration file has been created at {config_file}. Update the configuration in this file to match your system, and then re-run this program."
        )
        raise SystemExit(0)
    else:
        return config_file


def read_config_file(config_file: Path) -> dict:
    """
    Load the toml file at config_file and return the contents as a dictionary.

    Note: This function is called pre-logging setup, so any messages are printed to
    stdout.

    Args:
        config_file: path to the application configuration file

    Returns:
        config: configuration file contents as a dictionary
    """
    try:
        with open(config_file, "rb") as f:
            return tomllib.load(f)
    except Exception as err:
        print(f"ERROR: {err}")
        raise


def get_logdir(config: dict) -> Optional[Path]:
    """
    Return the path to the logging directory specified in config, or the default log
    directory if no path is specified.

    Args:
        config: restic-replica configuration dictionary

    Returns:
        logdir: path to logging directory specified in config
    """
    try:
        return Path(config["app"]["log_directory"]).expanduser()
    except KeyError:
        if platform.system() == "Windows":
            return Path.home() / "AppData/Local/restic-replica"
        else:
            return Path.home() / ".restic-replica"


def get_restic(config: dict, verbose: Optional[int] = 0) -> ResticCli:
    """
    Return a ResticCli instance populated with the information from config

    Args:
        config: restic configuration dictionary
        verbose: verbosity level of returned ResticCli instance

    Returns:
        a populated ResticCli instance
    """
    # if the restic path is not specified, set it
    try:
        path = Path(config["path"])
    except KeyError:
        if platform.system() == "Windows":
            path = Path("restic.exe")
        else:
            path = Path("restic")
    # environment variables
    try:
        env = config["environment"]
    except KeyError:
        env = {}
    # set RESTIC_PROGRESS_FPS if it isn't set
    if "RESTIC_PROGRESS_FPS" not in env.keys():
        env["RESTIC_PROGRESS_FPS"] = "0.016667"  # update every minute
    return ResticCli(path, environment_vars=env, verbose=verbose)


def get_policy(config: dict) -> Optional[Policy]:
    """
    Return a Policy instance populated with the information from config

    Args:
        config: policy configuration dictionary

    Returns:
        a populated Policy instance

    Raises:
        RuntimeError: raised if the policy is invalid, or invalid policy options are read from the config.
    """
    user_set_policy = False
    try:
        keep_last = config["keep-last"]
        user_set_policy = True
    except KeyError:
        keep_last = 0
    try:
        keep_daily = config["keep-daily"]
        user_set_policy = True
    except KeyError:
        keep_daily = 0
    try:
        keep_weekly = config["keep-weekly"]
        user_set_policy = True
    except KeyError:
        keep_weekly = 0
    try:
        keep_monthly = config["keep-monthly"]
        user_set_policy = True
    except KeyError:
        keep_monthly = 0
    try:
        keep_yearly = config["keep-yearly"]
        user_set_policy = True
    except KeyError:
        keep_yearly = 0
    try:
        no_current = config["exclude-current-period"]
    except KeyError:
        no_current = False

    # if user set any values, return a policy, else return none
    if user_set_policy:
        try:
            return Policy(
                keep_last,
                keep_daily,
                keep_weekly,
                keep_monthly,
                keep_yearly,
                no_current,
            )
        except (ValueError, TypeError) as err:
            raise RuntimeError(
                "Invalid policy; all keep-* options set in the config file must be non-negative integers, and at least one must be non-zero."
            ) from err
    else:
        return None


def get_repository(name: str, config: dict, restic_cli: ResticCli) -> Repository:
    """
    Return a Repository instance populated with the information from config

    Args:
        name: friendly name for the repsository
        config: repository configuration dictionary
        restic_cli: ResticCli instance that the repository will use for operations

    Returns:
        a populated Repository instance
    """
    # non-optional config data
    uri = config["repository_uri"]
    # optional config data
    try:
        password = config["password"]
    except KeyError:
        password = None
    try:
        password_file = config["password_file"]
    except KeyError:
        password_file = None
    try:
        password_command = config["password_command"]
    except KeyError:
        password_command = None
    try:
        env = config["environment"]
    except KeyError:
        env = None
    return Repository(
        uri,
        name,
        restic_cli,
        password=password,
        password_file=password_file,
        password_command=password_command,
        environment_vars=env,
    )


def check_repository_access(repository: Repository) -> bool:
    """
    Verify that a repository can be accessed successfully

    Args:
        repository: the Repository instance to check access for

    Returns:
        a boolean indicating operation success

    Raises:
        RunTimeError: raised if operation fails
    """
    try:
        return bool(repository.snapshots())
    except (CalledProcessError, OSError) as err:
        logger.error(err)
        raise RuntimeError(f"Unable to access restic repository {repository}") from err


def get_filtered_snapshots(repository: Repository, policy: Policy) -> SnapshotList:
    """"""
    snaps = SnapshotList.from_json(repository.snapshots(json=True).stdout)
    filtered_snaps = SnapshotList(snaps.filter(policy))
    if len(filtered_snaps.snapshots) > 0:
        return filtered_snaps
    else:
        raise RuntimeError("snapshot filtering led to 0 snapshots to copy")


def copy_snapshots(
    source_repository: Repository,
    destination_repository: Repository,
    policy: Optional[Policy] = None,
    dry_run: bool = False,
) -> CompletedProcess:
    """
    Copy snapshots from source_repository repository to destination_repository

    Args:
        source_repository: the Repository instance snapshots will be copied _from_
        destination_repository: the Repository instance snapshots will be copied _to_
        policy: an optional Policy instance that will be applied to filter the list of snapshots that will be copied
        dry_run: whether to actually perform the copy operation or not

    Returns:
        a populated CompletedProcess instance

    Raises:
        RunTimeError: raised if operation fails
        SystemExit: raised instead of performing the copy process, if dry_run is set
    """
    try:
        if policy:
            logger.info(
                f"Filtering snapshots to be copied from source repository using policy: {policy}"
            )
            filtered_snapshots = get_filtered_snapshots(source_repository, policy)
            logger.info(f"The following snapshots will be copied: {filtered_snapshots}")
        else:
            logger.info(
                "No policy specified, all snapshots will be copied from the source repository"
            )
            filtered_snapshots = None
        if dry_run:
            logger.info("dry-run flag set, exiting without performing copy operation")
            raise SystemExit(0)
        else:
            logger.info(
                f"Starting copy of snapshots from {source_repository.uri} to {destination_repository.uri}"
            )
            return destination_repository.copy(
                source_repository,
                live_output=True,
                snapshots=filtered_snapshots,
            )
    except (CalledProcessError, OSError) as err:
        logger.error(err)
        raise RuntimeError(
            f"error copying snapshots from {source_repository.uri} to {destination_repository.uri}"
        ) from err



from collections import namedtuple
from datetime import datetime
from dataclasses import dataclass
import json
from typing import Optional, Union


@dataclass
class Policy:
    """
    Rules for filtering a list of snapshots. Behaviour should approximate that of
    `restic forget`.

    All calendar related options (daily, weekly...) work on natural time boundaries, and
    are not relative to the time the Policy is applied. Weeks are Monday 00:00 to
    Sunday 23:59, days 00:00 to 23:59, etc. They also only count days/weeks/months/etc
    which have one or more snapshots.
    """

    last: int = 0
    daily: int = 0
    weekly: int = 0
    monthly: int = 0
    yearly: int = 0
    no_current: bool = False

    def __post_init__(self):
        if self.no_current not in (True, False):
            raise TypeError("exclude-current-period must be a boolean")
        for key in self.__dict__:
            # NB: this doesn't blow up because bool is a subclass of int.
            if not isinstance(self.__dict__[key], int):
                raise TypeError(f"{key} must be an integer")
            if self.__dict__[key] < 0:
                raise ValueError(f"{key} must be a non-negative integer")
        if sum(self.__dict__.values()) == 0:
            raise ValueError("Policy must contain at least one non-zero rule")

    def __str__(self):
        output = []
        if self.last > 0:
            output.append(f"keep-last={self.last}")
        if self.daily > 0:
            output.append(f"keep-daily={self.daily}")
        if self.weekly > 0:
            output.append(f"keep-weekly={self.weekly}")
        if self.monthly > 0:
            output.append(f"keep-monthly={self.monthly}")
        if self.yearly > 0:
            output.append(f"keep-yearly={self.yearly}")
        if self.no_current:
            output.append(f"exclude-current-period={self.no_current}")
        return ", ".join(output)


@dataclass
class Snapshot:
    """a restic snapshot"""

    time: datetime
    tree: str
    paths: list[str]
    hostname: str
    username: str
    uid: int
    gid: int
    program_version: str
    id: str
    short_id: str
    parent: Optional[str] = None  # An initial snapshot will have no parents
    original: Optional[str] = None  # present if a snapshot was rewritten
    # present if excludes were passed at backup time
    excludes: Optional[list[str]] = None
    tags: Optional[list[str]] = None  # present if snapshot has tags
    # summary was added in restic 0.17.0, so won't be included with snapshots created before then.
    summary: Optional[dict[str, Union[int | str]]] = None

    @classmethod
    def from_dict(cls, data):
        snap = cls(**data)
        # convert to timezone aware datetime object
        snap.time = datetime.fromisoformat(snap.time)
        return snap


@dataclass
class SnapshotList:
    """a list of restic snapshots"""

    snapshots: list[Snapshot]

    def __str__(self):
        return ", ".join([snap.short_id for snap in self.snapshots])

    @classmethod
    def from_json(cls, data):
        snap_list = []
        for snap in json.loads(data):
            snap_list.append(Snapshot.from_dict(snap))
        return cls(snap_list)

    def time_sorted(self, descending=False):
        """
        Return self.snapshots by the date the snapshot was taken

        Args:
            descending: whether snapshots are sorted in descending date order (newest
                first), or ascending order (oldest first). Defaults to False (ascending),
                this matches the ordering of `restic snapshots` output.

        Returns:
            self.snapshots, sorted by the date the snapshot was taken.
        """
        return sorted(self.snapshots, key=lambda s: s.time, reverse=descending)

    def filter(self, policy: Policy):
        """
        Filter self.snapshots according to a supplied Policy.

        Args:
            policy: Policy instance that defines rules for filtering a list of Snapshots

        Returns:
            filtered_snapshots: a filtered list of snapshots
        """
        # snapshots must be sorted in descending date order for filter methods to work correctly
        sorted_snapshots = self.time_sorted(descending=True)
        filtered_snapshots = []
        # store local timestamp to avoid timing inconsistencies between filters
        current_time = datetime.now()
        # Filters are additive. Start with least restrictive filters and work down from there.
        # NB: pass in sorted_snapshots to avoid repeatedly sorting the snapshot list
        # Apply the last filter
        if policy.last:
            filtered_snapshots.extend(self._filter_last(policy.last, sorted_snapshots))
        # Apply the daily filter
        if policy.daily:
            filtered_snapshots.extend(
                self._filter_daily(
                    policy.daily, sorted_snapshots, policy.no_current, current_time
                )
            )
        # Apply the weekly filter
        if policy.weekly:
            filtered_snapshots.extend(
                self._filter_weekly(
                    policy.weekly, sorted_snapshots, policy.no_current, current_time
                )
            )
        # Apply the monthly filter
        if policy.monthly:
            filtered_snapshots.extend(
                self._filter_monthly(
                    policy.monthly, sorted_snapshots, policy.no_current, current_time
                )
            )
        # Apply the yearly filter
        if policy.yearly:
            filtered_snapshots.extend(
                self._filter_yearly(
                    policy.yearly, sorted_snapshots, policy.no_current, current_time
                )
            )

        # remove duplicates, filtering by id as that is unique per-snapshot
        # store ids for snapshots that have already been seen in a set to eliminate duplicate ids
        seen_ids = set()
        # The "or seen_ids.add()" statement ensures each snapshot id is added to the set, it always returns None (which evaluates false-y).
        filtered_snapshots = [
            s
            for s in filtered_snapshots
            if not (s.id in seen_ids or seen_ids.add(s.id))
        ]
        # sort the filtered_snapshots by date _ascending_ (to match restic output)
        return sorted(filtered_snapshots, key=lambda s: s.time, reverse=False)

    def _filter_last(self, last: int, sorted_snapshots: list[Snapshot]):
        """
        Return the most recent 'n' snapshots from sorted_snapshots.

        Args:
            last: the number of snapshots to return
            sorted_snapshots: list of snapshots, sorted in descending date order.

        Returns:
            a filtered list of snapshots
        """
        return sorted_snapshots[:last]

    def _filter_daily(
        self,
        daily: int,
        sorted_snapshots: list[Snapshot],
        no_current: bool = False,
        current_time: Optional[datetime] = datetime.now(),
    ):
        """
        Return the most recent 'n' snapshots from sorted_snapshots, with a maximum of
        one snapshot per calendar day.
        Note: The most _recent_ snapshot from each caldendar day will be selected.

        Args:
            daily: the number of snapshots to return
            sorted_snapshots: list of snapshots, sorted in descending date order.
            no_current: whether snapshots within the current time period should be ignored.
            current_time: datetime aware object representing current time
        Returns:
            filtered_snapshots: a filtered list of snapshots
        """
        filtered_snapshots = []
        # datetime.isocalendar is unique on a per-day, per-year, and per-month basis.
        current_period = current_time.isocalendar()
        for snap in sorted_snapshots:
            # only process if target # of snapshots has not already been reached
            if len(filtered_snapshots) < daily:
                snap_period = snap.time.isocalendar()  # snapshot ts, to day resolution
                if not (no_current and snap_period == current_period):
                    if snap_period not in map(
                        lambda s: s.time.isocalendar(), filtered_snapshots
                    ):
                        filtered_snapshots.append(snap)
        return filtered_snapshots

    def _filter_weekly(
        self,
        weekly: int,
        sorted_snapshots: list[Snapshot],
        no_current: bool = False,
        current_time: Optional[datetime] = datetime.now(),
    ):
        """
        Return the most recent 'n' snapshots from sorted_snapshots, with a maximum of
        one snapshot per calendar week.
        Note: The most _recent_ snapshot from each caldendar week will be selected.

        Args:
            daily: the number of snapshots to return
            sorted_snapshots: list of snapshots, sorted in descending date order.
            no_current: whether snapshots within the current time period should be ignored.
            current_time: datetime aware object representing current time

        Returns:
            filtered_snapshots: a filtered list of snapshots
        """
        filtered_snapshots = []
        # use tuple that is unique on a per-week and per-year basis.
        YearWeek = namedtuple("YearWeek", "year week")
        # isocalendar().week gives week # in year
        current_period = YearWeek(current_time.year, current_time.isocalendar().week)
        for snap in sorted_snapshots:
            # only process if target # of snapshots has not already been reached
            if len(filtered_snapshots) < weekly:
                # snapshot ts, to week resolution
                snap_period = YearWeek(snap.time.year, snap.time.isocalendar().week)
                if not (no_current and snap_period == current_period):
                    if snap_period not in map(
                        lambda s: YearWeek(s.time.year, s.time.isocalendar().week),
                        filtered_snapshots,
                    ):
                        filtered_snapshots.append(snap)
        return filtered_snapshots

    def _filter_monthly(
        self,
        monthly: int,
        sorted_snapshots: list[Snapshot],
        no_current: bool = False,
        current_time: Optional[datetime] = datetime.now(),
    ):
        """
        Return the most recent 'n' snapshots from sorted_snapshots, with a maximum of
        one snapshot per calendar month.
        Note: The most _recent_ snapshot from each caldendar month will be selected.

        Args:
            daily: the number of snapshots to return
            sorted_snapshots: list of snapshots, sorted in descending date order.
            no_current: whether snapshots within the current time period should be ignored.
            current_time: datetime aware object representing current time

        Returns:
            filtered_snapshots: a filtered list of snapshots
        """
        filtered_snapshots = []
        # use tuple that is unique on a per-month and per-year basis.
        YearMonth = namedtuple("YearMonth", "year month")
        current_period = YearMonth(current_time.year, current_time.month)
        for snap in sorted_snapshots:
            # only process if target # of snapshots has not already been reached
            if len(filtered_snapshots) < monthly:
                # snapshot ts, to month resolution
                snap_period = YearMonth(snap.time.year, snap.time.month)
                if not (no_current and snap_period == current_period):
                    if snap_period not in map(
                        lambda s: YearMonth(s.time.year, s.time.month),
                        filtered_snapshots,
                    ):
                        filtered_snapshots.append(snap)
        return filtered_snapshots

    def _filter_yearly(
        self,
        yearly: int,
        sorted_snapshots: list[Snapshot],
        no_current: bool = False,
        current_time: Optional[datetime] = datetime.now(),
    ):
        """
        Return the most recent 'n' snapshots from sorted_snapshots, with a maximum of
        one snapshot per calendar year.
        Note: The most _recent_ snapshot from each caldendar year will be selected.

        Args:
            daily: the number of snapshots to return
            sorted_snapshots: list of snapshots, sorted in descending date order.
            no_current: whether snapshots within the current time period should be ignored.

        Returns:
            filtered_snapshots: a filtered list of snapshots
        """
        filtered_snapshots = []
        # filter only by year
        current_period = current_time.year
        for snap in sorted_snapshots:
            # only process if target # of snapshots has not already been reached
            if len(filtered_snapshots) < yearly:
                snap_period = snap.time.year
                if not (no_current and snap_period == current_period):
                    if snap_period not in map(
                        lambda s: s.time.year, filtered_snapshots
                    ):
                        filtered_snapshots.append(snap)
        return filtered_snapshots



import copy
from dataclasses import dataclass
import logging
import os
from pathlib import Path
import subprocess
from typing import Optional, Self

from restic_replica.snapshots import SnapshotList

logger = logging.getLogger(__name__)


@dataclass
class ResticCli:
    """
    The restic program interface

    Args:
        path: path to the restic binary
        environment_vars: dictionary of environment variables
        verbose: how verbose to be, max level/times is 2
    """

    path: Path
    environment_vars: dict[str, str]
    verbose: Optional[int] = 0

    def _execute_live_output(self, arguments: list[str]) -> subprocess.CompletedProcess:
        """
        Execute the command "arguments" and write stdout/stderr from the command to logger.
        Output will be writen out "line by line" as the process emits it.

        Args:
            arguments: the commandline arguments to be executed

        Returns:
           a subprocess.CompletedProcess object containing the exit code, stdout and
                stderr of the command.

        Raises:
            subprocess.CalledProcessError: raised if the command fails with an exit code
                that is not 0 or 3.
        """
        # use Popen instead of run to get "live" output
        with subprocess.Popen(
            arguments,
            stdout=subprocess.PIPE,
            stderr=subprocess.STDOUT,
            encoding="UTF-8",
            text=True,
        ) as process:
            line = None
            output = ""
            for line in process.stdout:
                if line is not None:
                    output = output + line
                    match line[0:5]:
                        case "Fatal":
                            logger.error(line.rstrip("\n"))
                        case "Warni":
                            logger.warning(line.rstrip("\n"))
                        case _:
                            logger.info(line.rstrip("\n"))
            process.wait()  # check for process termination
        # NB: restic returns code 3 if unable to read some source data during backup; this is only a partial failure.
        if process.returncode not in [0, 3]:
            raise subprocess.CalledProcessError(
                process.returncode, arguments, None, line
            )
        else:
            return subprocess.CompletedProcess(
                arguments, process.returncode, output, None
            )

    def execute(
        self,
        arguments: list[str],
        environment_vars: dict[
            str, str
        ] = {},  # NB: do not need to set to None, as the default value is only ever copied, never modified directly.
        live_output: bool = False,
        json: bool = False,
    ) -> subprocess.CompletedProcess:
        """
        Run the restic process with the provided commandline arguments.
        - If not live_output, stdout/stderr are captured and returned in a
        CompletedProcess instance after the restic process exits.
        - If live_output, stdout/stderr are written line-by-line to `logger` while the
        restic process is running, as well as being captured and returned in a
        CompletedProcess instance after the restic process exits.

        Args:
            arguments: the commandline arguments to be passed to restic.
            environment_vars: environment variables to pass to restic.
            live_output: whether to write output to `logger` as it is emitted by restic,
                or whether to capture output and return it after the restic process exits.
            json: whether restic should emit output in json format.

        Returns:
           a subprocess.CompletedProcess object containing the exit code, stdout and
                stderr of the command.
        """
        # ensure no mutation of mutable arguments
        local_args = copy.copy(arguments)
        local_env_vars = copy.copy(environment_vars)
        # add our environment variables to command, *overwriting duplicates*
        local_env_vars.update(self.environment_vars)
        # prepend restic binary path to args
        local_args.insert(0, str(self.path))
        # optionally set verbose level
        if self.verbose > 0:
            local_args.append(f"--verbose={self.verbose}")
        # optionally add json flag
        if json:
            local_args.append("--json")
        # set environment variables
        for key, value in local_env_vars.items():
            os.environ[key] = value

        try:
            if live_output:
                return self._execute_live_output(local_args)
            else:
                return subprocess.run(
                    local_args,
                    capture_output=True,
                    check=True,
                    encoding="UTF-8",
                    text=True,
                )
        finally:
            # unset all environment variables set previously
            for key in local_env_vars.keys():
                os.environ.pop(key, None)


@dataclass
class Repository:
    """restic repository class"""

    def __init__(
        self,
        uri: str,
        name: str,
        restic_cli: ResticCli,
        password: Optional[str] = None,
        password_file: Optional[str] = None,
        password_command: Optional[str] = None,
        environment_vars: Optional[dict] = None,
    ):
        self.uri = uri
        self.name = name
        self.restic_cli = restic_cli
        if environment_vars is None:
            self.environment_vars = {}
        else:
            self.environment_vars = environment_vars
        # any supplied passwords overwrite env vars, match restic behaviour of supplied cli arguments having precedence
        if password is not None:
            self.password = password
        if password_file is not None:
            self.password_file = password_file
        if password_command is not None:
            self.password_command = password_command
        # TODO: support for "--insecure-no-password" to be added here, also rclone.
        # validate a password has been supplied for the repository
        self._verify_password_is_set()

    @property
    def password(self) -> Optional[str]:
        try:
            return self.environment_vars["RESTIC_PASSWORD"]
        except KeyError:
            return None

    @password.setter
    def password(self, value: Optional[str]) -> None:
        self.environment_vars["RESTIC_PASSWORD"] = value

    @property
    def password_file(self) -> Optional[str]:
        try:
            return self.environment_vars["RESTIC_PASSWORD_FILE"]
        except KeyError:
            return None

    @password_file.setter
    def password_file(self, value: Optional[str]) -> None:
        self.environment_vars["RESTIC_PASSWORD_FILE"] = value

    @property
    def password_command(self) -> Optional[str]:
        try:
            return self.environment_vars["RESTIC_PASSWORD_COMMAND"]
        except KeyError:
            return None

    @password_command.setter
    def password_command(self, value: Optional[str]) -> None:
        self.environment_vars["RESTIC_PASSWORD_COMMAND"] = value

    def __str__(self):
        return self.uri

    def _common_args(self):
        return ["-r", f"{self.uri}"]

    def _verify_password_is_set(self) -> bool:
        """
        Assert that the password options supplied are valid. The behaviour of the restic
        program is mimiced here, in that:
        - password_file or password_command will overwrite password if set.
        - password_file and password_command are mutually exclusive; an error is raised
            if both are set.

        Returns:
            true if the options are valid.

        Raises:
            ValueError: raised if neither password, password_file or password_command
                are set.
            KeyError: raised if both password_file and password_command are set.
        """
        # check at least one password* option was supplied
        if not (self.password or self.password_file or self.password_command):
            raise ValueError(
                "one password argument is required: password, password_file, or password_command"
            )
        # check password_file and password_command were not both supplied.
        if self.password_file and self.password_command:
            raise KeyError("password_file and password_command are mutually exclusive")
        # print warning if password option will be overwritten; password-file and password-command take priority in restic
        if self.password and self.password_file:
            logger.warning(
                f"password and password_file were specified for repository {self.uri}; password_file will be used"
            )
        if self.password and self.password_command:
            logger.warning(
                f"password and password_command were specified for repository {self.uri}; password_command will be used"
            )
        return True

    def _filter_other_env(self, pair) -> bool:
        """
        Return false for incompatible environment variables, i.e. variables that can be
        combined with the environment variables of another Repository instance without
        causing authentication issues due to namespace collision.

        tldr; Filters out RESTIC_PASSWORD* env vars

        Args:
            pair: tuple consisting of dictionary key and value

        Returns:
            true to retain pair, false to discard
        """
        password_keys = [
            "RESTIC_PASSWORD",
            "RESTIC_PASSWORD_FILE",
            "RESTIC_PASSWORD_COMMAND",
        ]
        key, _ = pair  # unpack the pair
        if key in password_keys:
            return False
        else:
            return True

    def snapshots(
        self, live_output: bool = False, json: bool = False
    ) -> subprocess.CompletedProcess:
        """
        List the snapshots stored in this repository.

        Args:
            live_output: emit restic program output line-by-line
            json: set restic program output mode to json

        Returns:
           a subprocess.CompletedProcess object containing the exit code, stdout and
                stderr of the command.
        """

        # execute restic CLI with the snapshots argument
        args = self._common_args()
        args.extend(["snapshots"])
        return self.restic_cli.execute(
            args,
            environment_vars=self.environment_vars,
            live_output=live_output,
            json=json,
        )

    def copy(
        self,
        other: Self,
        live_output: bool = False,
        json: bool = False,
        snapshots: Optional[SnapshotList] = None,
    ) -> subprocess.CompletedProcess:
        """
        Copy snapshots from other repository to this repository.

        Args:
            other: respository to copy snapshots from
            live_output: emit restic program output line-by-line
            json: set restic program output mode to json
            snapshots: SnapshotList defining which snapshots to copy. If None, all snapshots will be copied.

        Returns:
           a subprocess.CompletedProcess object containing the exit code, stdout and
                stderr of the command.

        Raises:
            RuntimeError: raised if both source and destination repository are the same
                repository.
        """

        # prevent copying to/from the same repository
        if other.uri == self.uri:
            raise RuntimeError("source and destination repository must be different")

        # combine the environment variables from both repositories, giving primacy to our variables
        # NB: this is necessary to inherit S3 credentials and the like
        combined_env = dict(
            filter(self._filter_other_env, other.environment_vars.items())
        )
        # add our environment variables, overwriting any that already existed
        combined_env.update(self.environment_vars)

        # Set environment variables based on the password configuration of the other repository, in descending order of primacy
        if other.password_command:
            combined_env["RESTIC_FROM_PASSWORD_COMMAND"] = other.password_command
        elif other.password_file:
            combined_env["RESTIC_FROM_PASSWORD_FILE"] = other.password_file
        else:  # if neither a password command or password file were supplied, use password.
            combined_env["RESTIC_FROM_PASSWORD"] = other.password

        # execute restic CLI with the copy and other repository argument
        args = self._common_args()
        args.extend(["copy", "--from-repo", other.uri])

        # extend copy with list of snapshots
        if snapshots:
            for snapshot in snapshots.snapshots:
                args.append(snapshot.id)

        return self.restic_cli.execute(
            args,
            environment_vars=combined_env,
            live_output=live_output,
            json=json,
        )
